from amuse.community import *
from amuse.test.amusetest import TestWithMPI
from amuse import datamodel

from amuse.community.mpiamrvac.interface import MpiAmrVacInterface
from amuse.community.mpiamrvac.interface import MpiAmrVac
from amuse import io
from amuse.units import generic_unit_system

import os
import numpy


class TestMpiAmrVacInterface(TestWithMPI):

    def test1(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.initialize_code()
        instance.stop()

    def test2(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        filename, error = instance.get_parameters_filename()
        self.assertEqual(error, 0)
        self.assertEqual(filename, "amrvac.par")
        error = instance.set_parameters_filename("dontexists.file")
        self.assertEqual(error, -1)

        filename, error = instance.get_parameters_filename()
        self.assertEqual(error, 0)
        self.assertEqual(filename, "amrvac.par")

        name_of_parametersfile = 'amrvac.tst.par'
        with open(name_of_parametersfile, 'w') as f:
            f.write('test param')
        error = instance.set_parameters_filename(name_of_parametersfile)
        self.assertEqual(error, 0)

        filename, error = instance.get_parameters_filename()
        self.assertEqual(error, 0)
        self.assertEqual(filename, name_of_parametersfile)

        os.remove(name_of_parametersfile)

        instance.initialize_code()
        instance.stop()

    def test3(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.stop()

    def test4(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        error = instance.commit_parameters()
        self.assertEqual(error, 0)
        number_of_grids, error = instance.get_number_of_grids()
        self.assertEqual(error, 0)
        self.assertEqual(number_of_grids, 8)

        lx, ly, lz, error = instance.get_mesh_size(1)
        self.assertEqual(lx, 10)
        self.assertEqual(ly, 10)
        self.assertEqual(lz, 10)
        x, y, z, error = instance.get_position_of_index(0, 0, 0, 1)
        self.assertEqual(error, 0)
        self.assertEqual(x % 0.5, 0)
        self.assertEqual(y % 0.5, 0)
        self.assertEqual(z % 0.5, 0)
        instance.stop()
        # self.assertTrue(False)

    def test5(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        error = instance.commit_parameters()
        self.assertEqual(error, 0)

        number_of_grids, error = instance.get_number_of_grids()
        self.assertEqual(error, 0)
        self.assertEqual(number_of_grids, 8)

        rho, error = instance.get_grid_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0)
        error = instance.set_grid_density(1, 1, 1, 0.1, 1)
        self.assertEqual(error, 0)
        rho, error = instance.get_grid_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.1)
        rho, error = instance.get_grid_density(1, 1, 1, 2)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.0)

        instance.stop()

    def test6(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        error = instance.commit_parameters()
        self.assertEqual(error, 0)

        number_of_grids, error = instance.get_number_of_grids()
        self.assertEqual(error, 0)
        self.assertEqual(number_of_grids, 8)

        rho, error = instance.get_grid_energy_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0)
        error = instance.set_grid_energy_density(1, 1, 1, 0.1, 1)
        self.assertEqual(error, 0)
        rho, error = instance.get_grid_energy_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.1)
        rho, error = instance.get_grid_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.0)
        rho, error = instance.get_grid_energy_density(1, 1, 1, 2)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.0)
        instance.stop()

    def test7(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        error = instance.commit_parameters()
        self.assertEqual(error, 0)

        number_of_grids, error = instance.get_number_of_grids()
        self.assertEqual(error, 0)
        self.assertEqual(number_of_grids, 8)

        rhovx, rhovy, rhovz, error = instance.get_grid_momentum_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rhovx, 0.0)
        self.assertEqual(rhovy, 0.0)
        self.assertEqual(rhovz, 0.0)
        error = instance.set_grid_momentum_density(1, 1, 1, 0.1, 0.2, 0.3,  1)
        self.assertEqual(error, 0)
        rhovx, rhovy, rhovz, error = instance.get_grid_momentum_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rhovx, 0.1)
        self.assertEqual(rhovy, 0.2)
        self.assertEqual(rhovz, 0.3)
        rhovx, rhovy, rhovz, error = instance.get_grid_momentum_density(1, 1, 1, 2)
        self.assertEqual(error, 0)
        self.assertEqual(rhovx, 0.0)
        self.assertEqual(rhovy, 0.0)
        self.assertEqual(rhovz, 0.0)
        rho, error = instance.get_grid_energy_density(1, 1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(rho, 0.0)

        instance.stop()

    def test8(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        error = instance.set_boundary("periodic", "periodic", "periodic", "periodic", "periodic", "periodic")

        self.assertEqual(error, 0)
        error = instance.commit_parameters()
        self.assertEqual(error, 0)

        number_of_grids, error = instance.get_number_of_grids()
        self.assertEqual(error, 0)
        self.assertEqual(number_of_grids, 8)
        error = instance.initialize_grid()
        self.assertEqual(error, 0)

        instance.stop()

    def test9(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVacInterface)
        instance.set_parameters_filename(instance.default_parameters_filename)
        error = instance.initialize_code()
        self.assertEqual(error, 0)
        instance.setup_mesh(20, 20, 20, 20.0, 20.0, 20.0)
        instance.commit_parameters()
        n1, n2, n3, error = instance.get_acceleration_grid_size()
        self.assertEqual(error, 0)
        self.assertEqual(n1, 50)
        self.assertEqual(n2, 50)
        self.assertEqual(n3, 50)

        a1, a2, a3, error = instance.get_acceleration_grid_acceleration(1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(a1, 0.0)
        self.assertEqual(a2, 0)
        self.assertEqual(a3, 0)

        error = instance.set_acceleration_grid_acceleration(1, 1, 1, 10.0, 20.0, 30.0)
        self.assertEqual(error, 0)

        a1, a2, a3, error = instance.get_acceleration_grid_acceleration(1, 1, 1)
        self.assertEqual(error, 0)
        self.assertEqual(a1, 10.0)
        self.assertEqual(a2, 20.0)
        self.assertEqual(a3, 30.0)

        x, y, z, error = instance.get_acceleration_grid_position_of_index(1, 1, 1)
        self.assertEqual(error, 0)

        self.assertEqual(x, -1)
        self.assertEqual(y, -1)
        self.assertEqual(z, -1)
        instance.stop()


class TestMpiAmrVac(TestWithMPI):

    def test1(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.setup_mesh(20, 20, 20, 20.0 | generic_unit_system.length, 20.0 | generic_unit_system.length, 20.0 | generic_unit_system.length)
        error = instance.set_boundary("periodic", "periodic", "periodic", "periodic", "periodic", "periodic")

        rhovx, rhovy, rhovz = instance.get_grid_momentum_density(1, 1, 1, 1)

        self.assertEqual(rhovx, 0.0 | generic_unit_system.momentum_density)
        self.assertEqual(rhovy, 0.0 | generic_unit_system.momentum_density)
        self.assertEqual(rhovz, 0.0 | generic_unit_system.momentum_density)

        instance.stop()

    def test2(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 20)
        instance.set_boundary("periodic", "periodic", "periodic", "periodic", "periodic", "periodic")

        grids = list(instance.itergrids())

        self.assertEqual(len(grids), 8)

        for grid in grids:
            position = grids[0].position
            for i in range(3):
                max_x = position[..., i].amax()
                min_x = position[..., i].amin()

                self.assertTrue(min_x >= 0.5 | generic_unit_system.length)
                self.assertTrue(min_x <= 11.5 | generic_unit_system.length)
                self.assertTrue(max_x >= 9.5 | generic_unit_system.length)
                self.assertTrue(max_x <= 19.5 | generic_unit_system.length)

        self.assertEqual(grids[0][0][0][0].rho,  0.0 | generic_unit_system.density)

        grids[0].rho = 0.2 | generic_unit_system.density

        rho1 = grids[1].rho
        rho0 = grids[0].rho
        for i in range(10):
            for j in range(10):
                for k in range(10):
                    self.assertEqual(rho1[0][0][0],  0.0 | generic_unit_system.density)
                    self.assertEqual(rho0[0][0][0],  0.2 | generic_unit_system.density)

        instance.stop()

    def test3(self):

        for number_of_workers in range(2, 6):
            instance = self.new_instance_of_an_optional_code(MpiAmrVac, number_of_workers=number_of_workers)
            instance.set_parameters_filename(instance.default_parameters_filename)
            instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
            instance.parameters.mesh_size = (20, 20, 20)
            instance.parameters.x_boundary_conditions = ("periodic", "periodic")
            instance.parameters.y_boundary_conditions = ("periodic", "periodic")
            instance.parameters.z_boundary_conditions = ("periodic", "periodic")

            grids = list(instance.itergrids())

            self.assertEqual(len(grids), 8)

            for index, grid in enumerate(grids):
                position = grid.position
                # print instance.get_level_of_grid(index + 1)
                level = instance.get_level_of_grid(index + 1)
                self.assertEqual(level, 1)
                for i in range(3):
                    max_x = position[..., i].amax()
                    min_x = position[..., i].amin()

                    self.assertTrue(min_x >= 0.5 | generic_unit_system.length)
                    self.assertTrue(min_x <= 10.5 | generic_unit_system.length)
                    self.assertTrue(max_x >= 9.5 | generic_unit_system.length)
                    self.assertTrue(max_x <= 19.5 | generic_unit_system.length)
                    self.assertEqual(max_x - min_x, 9.0 | generic_unit_system.length)

            self.assertEqual(grids[0][0][0][0].rho,  0.0 | generic_unit_system.density)

            grids[0].rho = 0.2 | generic_unit_system.density

            rho1 = grids[4].rho
            rho0 = grids[0].rho
            for i in range(10):
                for j in range(10):
                    for k in range(10):
                        self.assertEqual(rho1[0][0][0],  0.0 | generic_unit_system.density)
                        self.assertEqual(rho0[0][0][0],  0.2 | generic_unit_system.density)

            instance.stop()

    def test4(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVac, number_of_workers=1)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 10.0) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.maximum_number_of_grid_levels = 5
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)

        levels = [instance.get_level_of_grid(i+1) for i in range(len(grids))]
        self.assertEqual(levels, [1] * 1)

        gamma = 5.0 / 3.0
        energy = generic_unit_system.mass / (generic_unit_system.time**2 * generic_unit_system.length)
        grids_in_memory = []
        for grid in grids:
            grid.rho = 1.0 | generic_unit_system.density
            grid.energy = (0.1795 | energy) / (gamma - 1)

        has_advanced = instance.refine_grid()
        grids = list(instance.itergrids())
        self.assertFalse(has_advanced)
        instance.stop()

    def test5(self):
        def fill_grids(grids):
            for grid in grids:
                firsthalf = grid.x > 5.0 | generic_unit_system.length
                secondhalf = numpy.logical_not(firsthalf)
                if(numpy.any(firsthalf)):
                    grid[firsthalf].rho = 1.0 | generic_unit_system.density

        instance = self.new_instance_of_an_optional_code(MpiAmrVac, number_of_workers=1)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 10.0) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.maximum_number_of_grid_levels = 4
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")

        n = 1
        for i in range(3):
            expected_length = n ** 3
            grids = list(instance.itergrids())
            levels = [instance.get_level_of_grid(grid+1) for grid in range(len(grids))]
            self.assertEqual(len(grids), expected_length)

            self.assertEqual(levels, [i+1] * expected_length)

            n *= 2

            fill_grids(grids)
            has_advanced = instance.refine_grid()
            self.assertTrue(has_advanced)

        has_advanced = instance.refine_grid()
        self.assertFalse(has_advanced)

        instance.stop()

    def test6(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac, number_of_workers=1)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 10.0) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")

        self.assertEqual(instance.parameters.maximum_number_of_grid_levels, 3)
        instance.parameters.maximum_number_of_grid_levels = 2
        self.assertEqual(instance.parameters.maximum_number_of_grid_levels, 2)

        self.assertEqual(instance.parameters.entropy_type, 'nul')
        instance.parameters.entropy_type = 'powell'
        self.assertEqual(instance.parameters.entropy_type, 'powell')

        self.assertEqual(instance.parameters.time_integration_procedure, 'twostep')
        instance.parameters.time_integration_procedure = 'onestep'
        self.assertEqual(instance.parameters.time_integration_procedure, 'onestep')

        self.assertEqual(instance.parameters.spatial_discretization_method, 'tvdmu')
        instance.parameters.spatial_discretization_method = 'tvdlf'
        self.assertEqual(instance.parameters.spatial_discretization_method, 'tvdlf')

        self.assertEqual(instance.parameters.predictor_step_discretization_method, 'tvdmu')
        instance.parameters.predictor_step_discretization_method = 'hancock'
        self.assertEqual(instance.parameters.predictor_step_discretization_method, 'hancock')

        instance.commit_parameters()

        self.assertEqual(instance.parameters.entropy_type, 'powell')
        self.assertEqual(instance.parameters.maximum_number_of_grid_levels, 2)
        self.assertEqual(instance.parameters.time_integration_procedure, 'onestep')
        self.assertEqual(instance.parameters.spatial_discretization_method, 'tvdlf')
        self.assertEqual(instance.parameters.predictor_step_discretization_method, 'hancock')

    def test7(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 10.0) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")

        grid = datamodel.Grid(10, 10, 10)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 | generic_unit_system.momentum_density
        grid.rhovz = 0.0 | generic_unit_system.momentum_density
        grid.energy = 1.0 | generic_unit_system.energy_density

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        instance.evolve_model(0.1 | generic_unit_system.time)

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.evolve_model(0.3 | generic_unit_system.time)
        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.stop()

    def test8(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.length_x = 1.0 | generic_unit_system.length
        instance.parameters.length_y = 1.0 | generic_unit_system.length
        instance.parameters.length_z = 1.0 | generic_unit_system.length
        instance.parameters.x_boundary_conditions = "periodic", "periodic"
        instance.parameters.y_boundary_conditions = "periodic", "periodic"
        instance.parameters.z_boundary_conditions = "periodic", "periodic"

        name = instance.get_typeghostfill()
        self.assertEqual(name, 'linear')
        instance.commit_parameters()

        name = instance.get_typeghostfill()
        self.assertEqual(name, 'linear')
        grid = datamodel.Grid(10, 10, 10)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 | generic_unit_system.momentum_density
        grid.rhovz = 0.0 | generic_unit_system.momentum_density
        grid.energy = 1.0 | generic_unit_system.energy_density

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        name = instance.get_typeghostfill()
        self.assertEqual(name, 'linear')

        self.assertEqual((50, 50, 50), instance.acceleration_grid.shape)

        acc_grid = datamodel.Grid(50, 50, 50)
        acceleration = 1 | generic_unit_system.acceleration
        acc_grid.ax = acceleration
        acc_grid.ay = acceleration
        acc_grid.az = acceleration
        # self.assertEquals(acc_grid.acceleration[0][0][0], ( 1,1,1) | generic_unit_system.acceleration)
        channel = acc_grid.new_channel_to(instance.acceleration_grid)
        channel.copy()

        name = instance.get_typeghostfill()
        self.assertEqual(name, 'linear')

        result = instance.initialize_grid()

        name = instance.get_typeghostfill()
        self.assertEqual(name, 'linear')

        instance.evolve_model(0.1 | generic_unit_system.time)

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)
        # self.assertAlmostRelativeEquals(igrid.rhovx, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);
        # self.assertAlmostRelativeEquals(igrid.rhovy, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);
        # self.assertAlmostRelativeEquals(igrid.rhovz, 0.1 * 1.0 * 0.1 | generic_unit_system.momentum_density);

        instance.evolve_model(0.3 | generic_unit_system.time)
        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)
        # self.assertAlmostRelativeEquals(igrid.rhovx, grid.rho *  instance.model_time * acceleration,2);
        # self.assertAlmostRelativeEquals(igrid.rhovy, grid.rho *  instance.model_time * acceleration,2);
        # self.assertAlmostRelativeEquals(igrid.rhovz, grid.rho *  instance.model_time * acceleration,2);
        instance.stop()

    def test9(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="2d")
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 10, 1)
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")

        grid = datamodel.Grid(10, 10, 1)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 | generic_unit_system.momentum_density
        grid.rhovz = 0.0 | generic_unit_system.momentum_density
        grid.energy = 1.0 | generic_unit_system.energy_density

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        instance.evolve_model(0.1 | generic_unit_system.time)

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.evolve_model(0.3 | generic_unit_system.time)
        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.stop()

    def test10(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_number_of_steps = 1

        gamma = 5.0 / 3.0

        grid = datamodel.new_regular_grid((10, 10, 10), [1.0, 1.0, 1.0] | generic_unit_system.length)

        density = generic_unit_system.density
        momentum = generic_unit_system.speed * generic_unit_system.density
        energy = generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)

        grid.rho = 0.01 | density
        grid.rhovx = 0.1 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.0 | momentum
        grid[0:5].rho = 0.015 | density

        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))

        grid.energy = p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2 + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(1.0 | generic_unit_system.time)

        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        rho = next(instance.itergrids()).rho[..., 0, 0]
        self.assertAlmostRelativeEquals(rho[7], 0.01 | density)
        self.assertTrue(rho[0] < 0.015 | density)
        self.assertTrue(instance.model_time < 1.0 | generic_unit_system.time)

        instance.stopping_conditions.number_of_steps_detection.disable()
        instance.evolve_model(0.1 | generic_unit_system.time)
        rho = next(instance.itergrids()).rho[..., 0, 0]
        self.assertAlmostRelativeEquals(instance.model_time, 0.1 | generic_unit_system.time)
        self.assertAlmostRelativeEquals(rho[7], 0.012812 | density, 3)
        self.assertTrue(rho[0] < 0.015 | density)

        instance.stop()

    def test11(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac)
        instance.parameters.mesh_size = (10, 10, 10)
        instance.parameters.mesh_length = [1.0, 1.0, 1.0] | generic_unit_system.length
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.stopping_conditions_timeout = 0.1 | units.s

        gamma = 5.0 / 3.0

        grid = datamodel.new_regular_grid((10, 10, 10), [1.0, 1.0, 1.0] | generic_unit_system.length)

        density = generic_unit_system.density
        momentum = generic_unit_system.speed * generic_unit_system.density
        energy = generic_unit_system.mass / ((generic_unit_system.time**2) * generic_unit_system.length)

        grid.rho = 0.01 | density
        grid.rhovx = 0.1 | momentum
        grid.rhovy = 0.0 | momentum
        grid.rhovz = 0.0 | momentum
        grid[0:5].rho = 0.015 | density

        p = 1.0 | (generic_unit_system.mass / (generic_unit_system.length * generic_unit_system.time**2))

        grid.energy = p / (gamma - 1)
        grid.energy += 0.5 * (grid.rhovx ** 2 + grid.rhovy ** 2 + grid.rhovz ** 2) / grid.rho

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        instance.stopping_conditions.timeout_detection.enable()
        instance.evolve_model(1.0 | generic_unit_system.time)

        self.assertTrue(instance.stopping_conditions.timeout_detection.is_set())
        rho = next(instance.itergrids()).rho[..., 0, 0]
        self.assertAlmostRelativeEquals(rho.mean(), 0.0125 | density)
        self.assertTrue(instance.model_time < 1.0 | generic_unit_system.time)

        instance.stopping_conditions.timeout_detection.disable()
        tnext = instance.model_time.round(2) + (0.2 | generic_unit_system.time)
        instance.evolve_model(tnext)
        rho = next(instance.itergrids()).rho[..., 0, 0]
        self.assertAlmostRelativeEquals(instance.model_time.round(2), tnext)
        self.assertAlmostRelativeEquals(rho.mean(), 0.0125 | density)
        self.assertTrue(rho[0] < 0.015 | density)

        instance.stop()

    def test12(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="1d")
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 1, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (10, 1, 1)
        instance.parameters.maximum_number_of_grid_levels = 1
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")

        grid = datamodel.Grid(10, 1, 1)
        grid.rho = 0.1 | generic_unit_system.density
        grid.rhovx = 0.0 | generic_unit_system.momentum_density
        grid.rhovy = 0.0 | generic_unit_system.momentum_density
        grid.rhovz = 0.0 | generic_unit_system.momentum_density
        grid.energy = 1.0 | generic_unit_system.energy_density

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        channel = grid.new_channel_to(igrid)
        channel.copy()

        instance.evolve_model(0.1 | generic_unit_system.time)

        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.evolve_model(0.3 | generic_unit_system.time)
        grids = list(instance.itergrids())
        self.assertEqual(len(grids), 1)
        igrid = grids[0]
        self.assertAlmostRelativeEquals(igrid.rho, grid.rho)

        instance.stop()

    def test13(self):
        instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="2d-acc")
        instance.set_parameters_filename(instance.default_parameters_filename)
        instance.parameters.mesh_length = (10.0, 10.0, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 1)
        instance.parameters.maximum_number_of_grid_levels = 2
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")

        must_refine = True
        middle = 5.0 | generic_unit_system.length
        while must_refine:
            must_refine = instance.refine_grid()

            for x in instance.itergrids():
                inmem = x.copy()
                inmem[inmem.x < middle].rho = 0.3 | generic_unit_system.density
                inmem[inmem.x >= middle].rho = 0.1 | generic_unit_system.density
                inmem.rhovx = 0.0 | generic_unit_system.momentum_density
                inmem.rhovy = 0.0 | generic_unit_system.momentum_density
                inmem.rhovz = 0.0 | generic_unit_system.momentum_density
                inmem.ax = 0.2 | generic_unit_system.acceleration
                inmem.ay = 0.0 | generic_unit_system.acceleration
                inmem.az = 0.0 | generic_unit_system.acceleration

                inmem.energy = 1.0 | generic_unit_system.energy_density
                from_model_to_code = inmem.new_channel_to(x)
                from_model_to_code.copy()
                self.assertAlmostRelativeEquals(x.ax, 0.2 | generic_unit_system.acceleration)
                self.assertAlmostRelativeEquals(x.ay, 0.0 | generic_unit_system.acceleration)
                self.assertAlmostRelativeEquals(x.az, 0.0 | generic_unit_system.acceleration)

        self.assertEqual(len(list(instance.itergrids())), 4)
        for igrid in instance.itergrids():
            self.assertAlmostRelativeEquals(igrid.ax, 0.2 | generic_unit_system.acceleration)
            self.assertAlmostRelativeEquals(igrid.ay, 0.0 | generic_unit_system.acceleration)
            self.assertAlmostRelativeEquals(igrid.az, 0.0 | generic_unit_system.acceleration)

        instance.evolve_model(0.1 | generic_unit_system.time)

        self.assertEqual(len(list(instance.itergrids())), 16)
        for igrid in instance.itergrids():
            self.assertAlmostRelativeEquals(igrid.ax, 0.2 | generic_unit_system.acceleration)
            self.assertAlmostRelativeEquals(igrid.ay, 0.0 | generic_unit_system.acceleration)
            self.assertAlmostRelativeEquals(igrid.az, 0.0 | generic_unit_system.acceleration)

        instance.stop()

    def test14(self):
        for ax, ay in ((0.2, 0.0), (0.0, 0.2), (0.2, 0.2)) | generic_unit_system.acceleration:
            instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="2d-acc")
            instance.set_parameters_filename(instance.default_parameters_filename)
            instance.parameters.mesh_length = (10.0, 10.0, 1) | generic_unit_system.length
            instance.parameters.mesh_size = (10, 10, 1)
            instance.parameters.maximum_number_of_grid_levels = 1
            instance.parameters.x_boundary_conditions = ("periodic", "periodic")
            instance.parameters.y_boundary_conditions = ("periodic", "periodic")
            rho = 0.1 | generic_unit_system.density
            middle = 5.0 | generic_unit_system.length
            for x in instance.itergrids():
                inmem = x.copy()
                inmem.rho = 0.1 | generic_unit_system.density
                inmem.rhovx = 0.0 | generic_unit_system.momentum_density
                inmem.rhovy = 0.0 | generic_unit_system.momentum_density
                inmem.ax = ax
                inmem.ay = ay

                inmem.energy = 1.0 | generic_unit_system.energy_density
                from_model_to_code = inmem.new_channel_to(x)
                from_model_to_code.copy()

            self.assertEqual(len(list(instance.itergrids())), 1)

            dt = 0.1 | generic_unit_system.time
            instance.evolve_model(dt)

            self.assertEqual(len(list(instance.itergrids())), 1)
            igrid = list(instance.itergrids())[0]

            self.assertAlmostRelativeEquals(igrid.rho, 0.1 | generic_unit_system.density)
            self.assertAlmostRelativeEquals(igrid.rhovx, ax * dt * rho)
            self.assertAlmostRelativeEquals(igrid.rhovy, ay * dt * rho)
            instance.stop()

    def test15(self):
        for ax, ay in ((0.2, 0.0), (0.0, 0.2), (0.2, 0.2)) | generic_unit_system.acceleration:  # (0.0,0.0),
            instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="2d-acc")
            instance.set_parameters_filename(instance.default_parameters_filename)
            instance.parameters.mesh_length = (10.0, 10.0, 1) | generic_unit_system.length
            instance.parameters.mesh_size = (20, 20, 1)
            instance.parameters.maximum_number_of_grid_levels = 1
            instance.parameters.time_accurate = False
            instance.parameters.x_boundary_conditions = ("periodic", "periodic")
            instance.parameters.y_boundary_conditions = ("periodic", "periodic")
            rho = 0.1 | generic_unit_system.density
            middle = 5.0 | generic_unit_system.length
            for x in instance.itergrids():
                inmem = x.copy()
                inmem.rho = 0.1 | generic_unit_system.density
                inmem.rhovx = 0.0 | generic_unit_system.momentum_density
                inmem.rhovy = 0.0 | generic_unit_system.momentum_density
                inmem.ax = ax
                inmem.ay = ay

                inmem.energy = 1.0 | generic_unit_system.energy_density
                from_model_to_code = inmem.new_channel_to(x)
                from_model_to_code.copy()

            self.assertEqual(len(list(instance.itergrids())), 4)

            dt = 0.1 | generic_unit_system.time
            instance.evolve_model(dt)
            dt = instance.model_time
            # ~ for igrid in instance.itergrids():
                # ~ print(ax * dt* rho, ax, instance.model_time)
            self.assertEqual(len(list(instance.itergrids())), 4)

            for igrid in instance.itergrids():
                # print igrid.rho[2]
                self.assertAlmostRelativeEquals(igrid.rho, rho)
                self.assertAlmostRelativeEquals(igrid.rhovx, ax * dt * rho)
                self.assertAlmostRelativeEquals(igrid.rhovy, ay * dt * rho)
            instance.stop()

    def test16(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="1d")
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.mesh_length = (20.0, 1, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 1, 1)
        instance.parameters.maximum_number_of_grid_levels = 1

        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = inmem.x/(1 | generic_unit_system.length) | generic_unit_system.density
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy = 1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5 | generic_unit_system.length, 0.0 | generic_unit_system.length, 0.0 | generic_unit_system.length)

        self.assertEqual(rho, 0.5 | generic_unit_system.density)

        for value in numpy.arange(0.5, 19.6, 0.1):

            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.0 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )

            self.assertAlmostRelativeEquals(rho, value | generic_unit_system.density)

        for value in numpy.arange(0.0, 0.6, 0.1):
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.0 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )
            self.assertAlmostRelativeEquals(rho, ((0.5 + value) * 0.5 + (0.5-value) * 19.5) | generic_unit_system.density)

        for value in numpy.arange(0.0, 0.5, 0.1):
            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value + 19.5 | generic_unit_system.length,
                0.0 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )
            self.assertAlmostRelativeEquals(rho, (19.5 - (value * 19)) | generic_unit_system.density, 9)

        # out of range
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
            20.0 | generic_unit_system.length,
            0.0 | generic_unit_system.length,
            0.0 | generic_unit_system.length
        )
        self.assertAlmostRelativeEquals(rho, 0.0 | generic_unit_system.density, 9)

    def test17(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVac, mode="2d", number_of_workers=2)
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.mesh_length = (20.0, 20.0, 1) | generic_unit_system.length
        instance.parameters.mesh_length = (20.0, 20.0, 1) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 1)
        instance.parameters.maximum_number_of_grid_levels = 1

        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = (inmem.x + ((inmem.y - (0.5 | generic_unit_system.length)) * 20.0))/(1 | generic_unit_system.length) | generic_unit_system.density
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy = 1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5 | generic_unit_system.length, 0.5 | generic_unit_system.length, 0.0 | generic_unit_system.length)

        self.assertEqual(rho, 0.5 | generic_unit_system.density)

        for value in numpy.arange(0.5, 19.6, 0.1):

            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.5 | generic_unit_system.length,
                0.0 | generic_unit_system.length
            )

            self.assertAlmostRelativeEquals(rho, value | generic_unit_system.density)

        for x in numpy.arange(8.5, 11.5, 0.25):
            for y in numpy.arange(0.5, 19.6, 0.25):
                rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                    x | generic_unit_system.length,
                    y | generic_unit_system.length,
                    0.0 | generic_unit_system.length
                )

                self.assertAlmostRelativeEquals(rho, x + (20 * (y-0.5)) | generic_unit_system.density)

    def test18(self):

        instance = self.new_instance_of_an_optional_code(MpiAmrVac, number_of_workers=3)
        instance.parameters.x_boundary_conditions = ("periodic", "periodic")
        instance.parameters.y_boundary_conditions = ("periodic", "periodic")
        instance.parameters.z_boundary_conditions = ("periodic", "periodic")
        instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
        instance.parameters.mesh_length = (20.0, 20.0, 20.0) | generic_unit_system.length
        instance.parameters.mesh_size = (20, 20, 20)
        instance.parameters.maximum_number_of_grid_levels = 1

        for x in instance.itergrids():
            inmem = x.copy()
            inmem.rho = (
                (
                    inmem.x +
                    ((inmem.y - (0.5 | generic_unit_system.length)) * 20.0) +
                    ((inmem.z - (0.5 | generic_unit_system.length)) * 400.0)
                )
                / (1 | generic_unit_system.length) | generic_unit_system.density
            )
            inmem.rhovx = 0.0 | generic_unit_system.momentum_density
            inmem.energy = 1.0 | generic_unit_system.energy_density
            from_model_to_code = inmem.new_channel_to(x)
            from_model_to_code.copy()
        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(0.5 | generic_unit_system.length, 0.5 | generic_unit_system.length, 0.5 | generic_unit_system.length)

        self.assertEqual(rho, 0.5 | generic_unit_system.density)

        for value in numpy.arange(0.5, 19.6, 0.1):

            rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
                value | generic_unit_system.length,
                0.5 | generic_unit_system.length,
                0.5 | generic_unit_system.length
            )

            self.assertAlmostRelativeEquals(rho, value | generic_unit_system.density)

        sample = sample = datamodel.new_regular_grid(
            (4, 4, 76),
            (2, 2, 19) | generic_unit_system.length
        )
        sample.x += 9.5 | generic_unit_system.length
        sample.y += 9.5 | generic_unit_system.length
        sample.z += 0.5 | generic_unit_system.length
        x = sample.x.flatten()
        y = sample.y.flatten()
        z = sample.z.flatten()

        rho, rhovx, rhovy, rhovx, rhoenergy = instance.get_hydro_state_at_point(
            x,
            y,
            z
        )
        half = 0.5 | generic_unit_system.length

        self.assertAlmostRelativeEquals(rho, (x + (20 * (y-half)) + (400 * (z-half)))/(1 | generic_unit_system.length) | generic_unit_system.density)



from amuse.support.interface import InCodeComponentImplementation

from amuse.community.distributed.interface import DistributedAmuse, Pilot

from amuse.test.amusetest import TestWithMPI
from amuse.support import exceptions

import os
import time
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi.tools import create_c
from amuse.rfi import channel
from amuse.rfi.core import *

from . import test_c_implementation


class TestCDistributedImplementationInterface(test_c_implementation.TestCImplementationInterface):

    @classmethod
    def setup_class(cls):
        cls.check_not_in_mpiexec()
        super(TestCDistributedImplementationInterface, cls).setup_class()
        # ~ print "Setting up distributed code"
        # instance = DistributedAmuse(redirection='none')
        cls.distinstance = cls.new_instance_of_an_optional_code(DistributedAmuse)  # , redirection='none')
        cls.distinstance.parameters.debug = False

        # ~ print "Resources:"
        # ~ print cls.distinstance.resources

        pilot = Pilot()
        pilot.resource_name = 'local'
        pilot.node_count = 1
        pilot.time = 2 | units.hour
        pilot.slots_per_node = 2
        pilot.label = 'local'
        cls.distinstance.pilots.add_pilot(pilot)
        # ~ print "Pilots:"
        # ~ print cls.distinstance.pilots

        # ~ print "Waiting for pilots"
        cls.distinstance.wait_for_pilots()
        cls.distinstance.use_for_all_workers()

    @classmethod
    def tearDown(cls):
        # ~ print "Stopping distributed code"
        cls.distinstance.stop()

    @classmethod
    def check_not_in_mpiexec(cls):
        """
        The tests will fork another process, if the test run
        is itself an mpi process, the tests may fail. 

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDI_CONTROL_FD' in os.environ:
            return
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            cls.skip('cannot run the socket tests under hydra process manager')

    def test22(self):
        self.skip("this test uses mpi internals, skip here")



from amuse.community.mobse.interface import MOBSE, MOBSEInterface

from amuse.test.amusetest import TestWithMPI
from amuse.units import units
from amuse.units import constants
from amuse.datamodel import Particles

import numpy


class TestMOBSEInterface(TestWithMPI):

    class state(object):
        def __init__(self):
            self.type1 = 0.0
            self.type2 = 0.0
            self.initial_mass1 = 0.0
            self.initial_mass2 = 0.0
            self.mass1 = 0.0
            self.mass2 = 0.0
            self.radius1 = 0.0
            self.radius2 = 0.0
            self.luminosity1 = 0.0
            self.luminosity2 = 0.0
            self.core_mass1 = 0.0
            self.core_mass2 = 0.0
            self.core_radius1 = 0.0
            self.core_radius2 = 0.0
            self.envelope_mass1 = 0.0
            self.envelope_mass2 = 0.0
            self.envelope_radius1 = 0.0
            self.envelope_radius2 = 0.0
            self.spin1 = 0.0
            self.spin2 = 0.0
            self.epoch1 = 0.0
            self.epoch2 = 0.0
            self.t_ms1 = 0.0
            self.t_ms2 = 0.0
            self.bse_age = 0.0
            self.orbital_period = 0.0
            self.eccentricity = 0.0

    def test1(self):
        print("Test initialization...")
        instance = MOBSEInterface()
        metallicity = 0.02
        neta = 0.5
        bwind = 0.0
        hewind = 0.5
        alpha1 = 1.0
        CElambda = 0.1
        ceflag = 0
        tflag = 1
        ifflag = 0
        wdflag = 1
        bhflag = 1
        nsflag = 3
        piflag = 1
        mxns = 3.0
        idum = 29769
        pts1 = 0.05
        pts2 = 0.01
        pts3 = 0.02
        sigma1 = 265.0
        sigma2 = 7.0
        beta = 1.0/8.0
        xi = 1.0
        acc2 = 3.0/2.0
        epsnov = 0.001
        eddfac = 1.0
        gamma = -1.0

        status = instance.initialize(metallicity,
            neta, bwind, hewind, alpha1, CElambda,
            ceflag, tflag, ifflag, wdflag, bhflag,
            nsflag, piflag, mxns, idum, pts1, pts2, pts3,
            sigma1, sigma2, beta, xi, acc2, epsnov, eddfac, gamma)
        self.assertEqual(status, 0)
        instance.stop()

    def test2(self):
        print("Test basic operations (legacy functions evolve & get_time_step)...")
        instance = MOBSEInterface()
        status = instance.initialize(0.02, 0.5, 0.0, 0.5, 1.0, 0.1, 0, 1, 0, 1, 1, 3, 1, 3.0,
            29769, 0.05, 0.01, 0.02, 265.0, 7.0, 1.0/8.0, 1.0, 3.0/2.0, 0.001, 1.0, -1.0)

        new_state = self.state()
        new_state.mass1 = 3.0
        new_state.mass2 = 1.0
        new_state.initial_mass1 = 3.0
        new_state.initial_mass2 = 1.0
        new_state.type1 = 1.0
        new_state.type2 = 1.0
        new_state.end_time = 1e-06
        new_state.orbital_period = 200.0
        new_state.eccentricity = 0.5

        result = instance.evolve_binary(
            new_state.type1, new_state.type2, new_state.initial_mass1, new_state.initial_mass2,
            new_state.mass1, new_state.mass2, new_state.radius1, new_state.radius2,
            new_state.luminosity1, new_state.luminosity2, new_state.core_mass1,
            new_state.core_mass2, new_state.core_radius1, new_state.core_radius2,
            new_state.envelope_mass1, new_state.envelope_mass2, new_state.envelope_radius1,
            new_state.envelope_radius2, new_state.spin1, new_state.spin2, new_state.epoch1,
            new_state.epoch2, new_state.t_ms1, new_state.t_ms2, new_state.bse_age,
            new_state.orbital_period, new_state.eccentricity, new_state.end_time
        )

        updated_state = self.state()
        (updated_state.type1, updated_state.type2, updated_state.initial_mass1, updated_state.initial_mass2,
            updated_state.mass1, updated_state.mass2, updated_state.radius1, updated_state.radius2,
            updated_state.luminosity1, updated_state.luminosity2, updated_state.core_mass1,
            updated_state.core_mass2, updated_state.core_radius1, updated_state.core_radius2,
            updated_state.envelope_mass1, updated_state.envelope_mass2, updated_state.envelope_radius1,
            updated_state.envelope_radius2, updated_state.spin1, updated_state.spin2,
            updated_state.epoch1, updated_state.epoch2, updated_state.t_ms1, updated_state.t_ms2,
            updated_state.bse_age, updated_state.orbital_period,
            updated_state.eccentricity, updated_state.end_time) = result

        expected = {
            'radius2': '0x1.c6c8a1c793bcep-1',
            'luminosity2': '0x1.653b1b2d0333bp-1',
            'core_mass2': '0x0.0p+0',
            'bse_age': '0x1.0c6f7a0b5ed8dp-20',
            'end_time': '0x1.0c6f7a0b5ed8dp-20',
            'envelope_mass2': '0x1.0d6fc100ab510p-5',
            'mass2': '0x1.0000000000000p+0',
            'initial_mass2': '0x1.0000000000000p+0',
            'envelope_radius2': '0x1.db27631ba0e5ap-3',
            'core_radius2': '0x0.0p+0',
            'epoch2': '0x0.0p+0',
            't_ms2': '0x1.57d90abe54643p+13',
            'spin2': '0x1.07413b0522dabp+10',
        }

        for x in expected:
            print("'%s' : '%s'," % (x, getattr(updated_state, x).hex()))
            self.assertAlmostRelativeEqual(float.fromhex(expected[x]), getattr(updated_state, x))

        self.assertEqual(updated_state.end_time, 1e-06)
        dt = instance.get_time_step(updated_state.type1, updated_state.type2,
            updated_state.initial_mass1, updated_state.initial_mass2, updated_state.mass1,
            updated_state.mass2, updated_state.t_ms1, updated_state.t_ms2,
            updated_state.epoch1, updated_state.epoch2, updated_state.bse_age)
        self.assertAlmostEqual(dt, 18.8768, 3)
        instance.stop()

    def test3(self):
        print("Test whether the interface can handle arrays...")
        instance = MOBSEInterface()
        status = instance.initialize(0.02, 0.5, 0.0, 0.5, 1.0, 0.1, 0, 1, 0, 1, 1, 3, 1, 3.0,
            29769, 0.05, 0.01, 0.02, 265.0, 7.0, 1.0/8.0, 1.0, 3.0/2.0, 0.001, 1.0, -1.0)
        masses1 = [10.0, 5.0, 4.0, 40.0, 130.0]
        masses2 = [1.0, 1.0, 1.0, 10.0, 60.0]
        types1 = types2 = [1, 1, 1, 1, 1]
        orbital_periods = [100.0, 200.0, 300.0, 500.0, 600.0]
        eccentricities = [0.5, 0.6, 0.7, 0.3, 0.2]

        radii1 = luminosity1 = core_mass1 = core_radius1 = envelope_mass1 =\
        envelope_radius1 = spin1 = epoch1 = t_ms1 = [0.0, 0.0, 0.0, 0.0, 0.0]
        radii2 = luminosity2 = core_mass2 = core_radius2 = envelope_mass2 =\
        envelope_radius2 = spin2 = epoch2 = t_ms2 = [0.0, 0.0, 0.0, 0.0, 0.0]
        init_mass1 = masses1
        init_mass2 = masses2
        bse_age = [0.0, 0.0, 0.0, 0.0, 0.0]
        end_time = [10.0, 10.0, 10.0, 10.0, 10.0]
        result = instance.evolve_binary(
            types1, types2, init_mass1, init_mass2,
            masses1, masses2, radii1, radii2,
            luminosity1, luminosity2, core_mass1, core_mass2,
            core_radius1, core_radius2, envelope_mass1, envelope_mass2,
            envelope_radius1, envelope_radius2, spin1, spin2,
            epoch1, epoch2, t_ms1, t_ms2,
            bse_age, orbital_periods, eccentricities, end_time
        )
        self.assertAlmostEqual(result['mass1'][0], 9.99356, 2)
        self.assertAlmostEqual(result['mass1'][1], 4.99956, 2)
        self.assertAlmostEqual(result['mass1'][2], 3.99992, 2)
        self.assertAlmostEqual(result['mass1'][3], 3.07374, 2)
        self.assertAlmostEqual(result['mass1'][4], 13.61644, 2)
        instance.stop()

    def test4(self):
        print("Test large number of particles...")
        number_of_particles = 2000
        instance = MOBSEInterface()
        status = instance.initialize(0.02, 0.5, 0.0, 0.5, 1.0, 0.1, 0, 1, 0, 1, 1, 3, 1, 3.0,
            29769, 0.05, 0.01, 0.02, 265.0, 7.0, 1.0/8.0, 1.0, 3.0/2.0, 0.001, 1.0, -1.0)
        masses1 = [1.0 + ((x / 1.0*number_of_particles) * 10.0) for x in range(1, number_of_particles+1)]
        masses2 = [2.0 + ((x / 1.0*number_of_particles) * 5.0) for x in range(1, number_of_particles+1)]
        orbital_periods = [100.0 + ((x / 1.0*number_of_particles) * 900.0) for x in range(1, number_of_particles+1)]
        eccentricities = [0.5 + ((x / 1.0*number_of_particles) * 0.4) for x in range(1, number_of_particles+1)]

        types1 = types2 = [1 for x in range(1, number_of_particles+1)]
        radii1 = luminosity1 = core_mass1 = core_radius1 = envelope_mass1 =\
        envelope_radius1 = spin1 = epoch1 = t_ms1 =\
        radii2 = luminosity2 = core_mass2 = core_radius2 = envelope_mass2 =\
        envelope_radius2 = spin2 = epoch2 = t_ms2 =\
        bse_age = [0.0 for x in range(1, number_of_particles+1)]
        end_time = [1.0 for x in range(1, number_of_particles+1)]
        init_mass1 = masses1
        init_mass2 = masses2

        result = instance.evolve_binary(
            types1, types2, init_mass1, init_mass2,
            masses1, masses2, radii1, radii2,
            luminosity1, luminosity2, core_mass1, core_mass2,
            core_radius1, core_radius2, envelope_mass1, envelope_mass2,
            envelope_radius1, envelope_radius2, spin1, spin2,
            epoch1, epoch2, t_ms1, t_ms2,
            bse_age, orbital_periods, eccentricities, end_time
        )
        self.assertEqual(len(result['mass1']), number_of_particles)
        instance.stop()


class TestMOBSE(TestWithMPI):

    def test1(self):
        print("Testing evolution of a close binary system...")
        instance = MOBSE()
        instance.initialize_code()
        instance.parameters.metallicity = 0.001
        instance.parameters.common_envelope_efficiency = 3.0
        instance.parameters.Eddington_mass_transfer_limit_factor = 10.0
        instance.commit_parameters()
        stars = Particles(2)
        stars[0].mass = 3.0 | units.MSun
        stars[1].mass = 0.3 | units.MSun

        orbital_period = 200.0 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)

        instance.particles.add_particles(stars)

        binaries = Particles(1)

        binary = binaries[0]
        binary.semi_major_axis = semi_major_axis
        binary.eccentricity = 0.5
        binary.child1 = stars[0]
        binary.child2 = stars[1]

        instance.binaries.add_particles(binaries)

        from_mobse_to_model = instance.particles.new_channel_to(stars)
        from_mobse_to_model.copy()

        from_mobse_to_model_binaries = instance.binaries.new_channel_to(binaries)
        from_mobse_to_model_binaries.copy()

        previous_type = binary.child1.stellar_type
        results = []
        current_time = 0 | units.Myr

        while current_time < (480 | units.Myr):
            instance.update_time_steps()
            # The next line appears a bit weird, but saves time for this simple test.
            current_time = current_time + max(5.0*instance.binaries[0].time_step, 0.3 | units.Myr)
            instance.evolve_model(current_time)
            from_mobse_to_model.copy()
            from_mobse_to_model_binaries.copy()
            if not binary.child1.stellar_type == previous_type:
                results.append((binary.age, binary.child1.mass, binary.child1.stellar_type))
                previous_type = binary.child1.stellar_type

        self.assertEqual(len(results), 6)

        types = (
            "Hertzsprung Gap",
            "First Giant Branch",
            "Core Helium Burning",
            "First Asymptotic Giant Branch",
            "Hertzsprung Gap Naked Helium star",
            "Carbon/Oxygen White Dwarf",
        )

        for result, expected in zip(results, types):
            self.assertEqual(str(result[2]), expected)

        times = (
            284.8632 | units.Myr,
            287.0713 | units.Myr,
            287.7967 | units.Myr,
            331.1631 | units.Myr,
            331.4164 | units.Myr,
            332.2864 | units.Myr,
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(result[0].value_in(units.Myr), expected.value_in(units.Myr), 0)

        masses = (
            3.000 | units.MSun,
            3.000 | units.MSun,
            2.999 | units.MSun,
            2.956 | units.MSun,
            0.888 | units.MSun,
            0.701 | units.MSun,
        )
        for result, expected in zip(results, masses):
            self.assertAlmostEqual(result[1].value_in(units.MSun), expected.value_in(units.MSun), 2)

        instance.stop()

    def test2(self):
        print("Testing evolution of a wide binary system.")
        instance = MOBSE()
        instance.parameters.metallicity = 0.001
        instance.parameters.common_envelope_efficiency = 3.0
        instance.parameters.Eddington_mass_transfer_limit_factor = 10.0
        instance.commit_parameters()

        stars = Particles(2)
        stars[0].mass = 4.0 | units.MSun
        stars[1].mass = 0.3 | units.MSun
        orbital_period = 2.0e5 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)

        instance.particles.add_particles(stars)

        binaries = Particles(1)

        binary = binaries[0]
        binary.semi_major_axis = semi_major_axis
        binary.eccentricity = 0.5
        binary.child1 = stars[0]
        binary.child2 = stars[1]

        instance.binaries.add_particles(binaries)

        from_mobse_to_model = instance.particles.new_channel_to(stars)
        from_mobse_to_model.copy()

        from_mobse_to_model_binaries = instance.binaries.new_channel_to(binaries)
        from_mobse_to_model_binaries.copy()

        previous_type = binary.child1.stellar_type
        results = []
        current_time = 0 | units.Myr

        while current_time < (170 | units.Myr):
            instance.update_time_steps()
            # The next line appears a bit weird, but saves time for this simple test.
            current_time = current_time + max(2.0*instance.binaries[0].time_step, 0.04 | units.Myr)
            instance.evolve_model(current_time)
            from_mobse_to_model.copy()
            from_mobse_to_model_binaries.copy()
            if not binary.child1.stellar_type == previous_type:
                results.append((binary.age, binary.child1.mass, binary.child1.stellar_type))
                previous_type = binary.child1.stellar_type
        print(results)
        self.assertEqual(len(results), 6)

        times = (
            147.1282 | units.Myr,
            148.0345 | units.Myr,
            148.2282 | units.Myr,
            167.2811 | units.Myr,
            168.0344 | units.Myr,
            168.7475 | units.Myr
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(result[0].value_in(units.Myr), expected.value_in(units.Myr), 0)

        masses = (
            4.000 | units.MSun,
            3.999 | units.MSun,
            3.999 | units.MSun,
            3.942 | units.MSun,
            3.906 | units.MSun,
            1.016 | units.MSun
        )
        for result, expected in zip(results, masses):
            self.assertAlmostEqual(result[1].value_in(units.MSun), expected.value_in(units.MSun), 2)

        types = (
            "Hertzsprung Gap",
            "First Giant Branch",
            "Core Helium Burning",
            "First Asymptotic Giant Branch",
            "Second Asymptotic Giant Branch",
            "Carbon/Oxygen White Dwarf",
        )

        for result, expected in zip(results, types):
            self.assertEqual(str(result[2]), expected)

        instance.stop()

    def test3(self):
        print("Testing standard MOBSE example 2...")
        instance = MOBSE()
        instance.parameters.common_envelope_efficiency = 3.0
        instance.parameters.common_envelope_binding_energy_factor = 0.5
        instance.parameters.Eddington_mass_transfer_limit_factor = 10.0
        instance.commit_parameters()

        stars = Particles(2)
        stars[0].mass = 7.816 | units.MSun
        stars[1].mass = 4.387 | units.MSun

        orbital_period = 1964.18453 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)
        instance.particles.add_particles(stars)

        binaries = Particles(1)

        binary = binaries[0]
        binary.semi_major_axis = semi_major_axis
        binary.eccentricity = 0.0
        binary.child1 = stars[0]
        binary.child2 = stars[1]

        instance.binaries.add_particles(binaries)

        from_mobse_to_model = instance.particles.new_channel_to(stars)
        from_mobse_to_model.copy()

        from_mobse_to_model_binaries = instance.binaries.new_channel_to(binaries)
        from_mobse_to_model_binaries.copy()

        previous_type1 = binary.child1.stellar_type
        previous_type2 = binary.child2.stellar_type
        results = []
        current_time = 0 | units.Myr

        while current_time < (170 | units.Myr):
            instance.update_time_steps()
            # The next line appears a bit weird, but saves time for this simple test.
            current_time = current_time + max(2.0*instance.binaries[0].time_step, 0.04 | units.Myr)
            instance.evolve_model(current_time)
            from_mobse_to_model.copy()
            from_mobse_to_model_binaries.copy()
            if not (binary.child1.stellar_type == previous_type1 and binary.child2.stellar_type == previous_type2):
                results.append((binary.age, str(binary.child1.stellar_type)+" and "+str(binary.child2.stellar_type)))
                previous_type1 = binary.child1.stellar_type
                previous_type2 = binary.child2.stellar_type

        print('\n'.join(map(str, results)))
        self.assertEqual(len(results), 12)
        times = (
            39.1037 | units.Myr,
            39.2242 | units.Myr,
            39.2565 | units.Myr,
            43.9911 | units.Myr,
            44.1842 | units.Myr,
            44.2644 | units.Myr,
            141.8444 | units.Myr,
            142.4835 | units.Myr,
            142.9234 | units.Myr,
            166.3238 | units.Myr,
            166.8385 | units.Myr,
            167.1731 | units.Myr
        )
        for result, expected in zip(results, times):
            self.assertAlmostEqual(result[0].value_in(units.Myr), expected.value_in(units.Myr), 0)

        types = (
            "Hertzsprung Gap and Main Sequence star",
            "First Giant Branch and Main Sequence star",
            "Core Helium Burning and Main Sequence star",
            "First Asymptotic Giant Branch and Main Sequence star",
            "Second Asymptotic Giant Branch and Main Sequence star",
            "Oxygen/Neon White Dwarf and Main Sequence star",
            "Oxygen/Neon White Dwarf and Hertzsprung Gap",
            "Oxygen/Neon White Dwarf and First Giant Branch",
            "Oxygen/Neon White Dwarf and Core Helium Burning",
            "Oxygen/Neon White Dwarf and First Asymptotic Giant Branch",
            "Oxygen/Neon White Dwarf and Hertzsprung Gap Naked Helium star",
            "Neutron Star and Carbon/Oxygen White Dwarf",
        )

        for result, expected in zip(results, types):
            self.assertEqual(result[1], expected)

        self.assertAlmostEqual(binary.child1.mass.value_in(units.MSun), 1.26079, 3)
        self.assertAlmostEqual(binary.child2.mass.value_in(units.MSun), 0.76080, 3)

        instance.stop()

    def test4(self):
        print("Quick testing standard MOBSE example 2...")
        instance = MOBSE()
        instance.parameters.common_envelope_efficiency = 3.0
        instance.parameters.common_envelope_binding_energy_factor = 0.5
        instance.parameters.Eddington_mass_transfer_limit_factor = 10.0
        instance.commit_parameters()

        stars = Particles(2)
        stars[0].mass = 7.816 | units.MSun
        stars[1].mass = 4.387 | units.MSun

        instance.particles.add_particles(stars)

        binaries = Particles(1)

        binary = binaries[0]
        orbital_period = 1964.18453 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)
        binary.semi_major_axis = semi_major_axis
        binary.eccentricity = 0.0
        binary.child1 = stars[0]
        binary.child2 = stars[1]

        instance.binaries.add_particles(binaries)

        from_mobse_to_model = instance.particles.new_channel_to(stars)
        from_mobse_to_model.copy()

        from_mobse_to_model_binaries = instance.binaries.new_channel_to(binaries)
        from_mobse_to_model_binaries.copy()

        instance.evolve_model(170 | units.Myr)
        from_mobse_to_model.copy()
        from_mobse_to_model_binaries.copy()

        self.assertAlmostEqual(binary.child1.mass.value_in(units.MSun), 1.26079, 3)
        self.assertAlmostEqual(binary.child2.mass.value_in(units.MSun), 0.76080, 3)
        self.assertEqual(str(binary.child1.stellar_type), "Neutron Star")
        self.assertEqual(str(binary.child2.stellar_type), "Carbon/Oxygen White Dwarf")

        instance.stop()

    def test5(self):
        print("Testing stellar collision...")
        instance = MOBSE()
        instance.parameters.common_envelope_efficiency = 3.0
        instance.parameters.Eddington_mass_transfer_limit_factor = 10.0
        instance.commit_parameters()

        stars = Particles(2)
        stars[0].mass = 130.0 | units.MSun
        stars[1].mass = 50 | units.MSun

        instance.particles.add_particles(stars)

        binaries = Particles(1)

        binary = binaries[0]
        orbital_period = 300.0 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)
        binary.semi_major_axis = semi_major_axis
        binary.eccentricity = 0.99
        binary.child1 = stars[0]
        binary.child2 = stars[1]

        instance.binaries.add_particles(binaries)

        from_mobse_to_model = instance.particles.new_channel_to(stars)
        from_mobse_to_model.copy()

        from_mobse_to_model_binaries = instance.binaries.new_channel_to(binaries)
        from_mobse_to_model_binaries.copy()

        instance.evolve_model(170 | units.Myr)

        from_mobse_to_model.copy()
        from_mobse_to_model_binaries.copy()
        print(binaries)
        self.assertAlmostEqual(binary.child1.mass.value_in(units.MSun), 180.00, 3)
        self.assertAlmostEqual(binary.child2.mass.value_in(units.MSun), 0.000, 3)
        self.assertEqual(str(binary.child1.stellar_type), "Main Sequence star")
        self.assertEqual(str(binary.child2.stellar_type), "Massless Supernova")

        instance.stop()

    def test6(self):
        print("Testing additional parameters for initialization...")
        instance = MOBSE()
        instance.initialize_code()
        self.assertEqual(instance.parameters.reimers_mass_loss_coefficient, 0.5)
        myvalue = 0.7
        instance.parameters.reimers_mass_loss_coefficient = myvalue
        self.assertEqual(instance.parameters.reimers_mass_loss_coefficient, myvalue)
        instance.commit_parameters()
        self.assertEqual(instance.parameters.reimers_mass_loss_coefficient, myvalue)
        instance.stop()

        instance = MOBSE()
        self.assertEqual(instance.parameters.reimers_mass_loss_coefficient, 0.5)
        myvalue = 0.7
        instance.parameters.reimers_mass_loss_coefficient = myvalue
        instance.parameters.set_defaults()
        instance.commit_parameters()
        self.assertEqual(instance.parameters.reimers_mass_loss_coefficient, 0.5)
        instance.stop()

    def test7(self):
        print("Test evolve_model optional arguments: end_time and keep_synchronous")

        instance = MOBSE()
        instance.commit_parameters()

        stars = Particles(6)
        stars.mass = [1.0, 2.0, 3.0, 0.1, 0.2, 0.3] | units.MSun

        binaries = Particles(3)
        binaries.eccentricity = 0.0
        for i in range(3):
            binaries[i].child1 = stars[i]
            binaries[i].child2 = stars[i+3]
        orbital_period = 200.0 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(
            orbital_period,
            binaries.child1.as_set().mass,
            binaries.child2.as_set().mass
        )
        binaries.semi_major_axis = semi_major_axis

        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binaries)

        self.assertAlmostEqual(instance.binaries.age, [0.0, 0.0, 0.0] | units.yr)
        self.assertAlmostEqual(instance.binaries.time_step, [550.1565, 58.2081, 18.8768] | units.Myr, 3)

        print("evolve_model without arguments: use shared timestep = min(particles.time_step)")
        instance.evolve_model()
        self.assertAlmostEqual(instance.binaries.age, [18.8768, 18.8768, 18.8768] | units.Myr, 3)
        self.assertAlmostEqual(instance.binaries.time_step, [550.1565, 58.2081, 18.8768] | units.Myr, 3)
        self.assertAlmostEqual(instance.model_time, 18.8768 | units.Myr, 3)

        print("evolve_model with end_time: take timesteps, until end_time is reached exactly")
        instance.evolve_model(100 | units.Myr)
        self.assertAlmostEqual(instance.binaries.age, [100.0, 100.0, 100.0] | units.Myr, 3)
        self.assertAlmostEqual(instance.binaries.time_step, [550.1565, 58.2081, 18.8785] | units.Myr, 3)
        self.assertAlmostEqual(instance.model_time, 100.0 | units.Myr, 3)

        print("evolve_model with keep_synchronous: use non-shared timestep, particle ages will typically diverge")
        instance.evolve_model(keep_synchronous=False)
        self.assertAlmostEqual(instance.binaries.age, (100 | units.Myr) + ([550.1565, 58.2081, 18.8785] | units.Myr), 3)
        self.assertAlmostEqual(instance.binaries.time_step, [550.1565, 58.2081, 18.8785] | units.Myr, 3)
        self.assertAlmostEqual(instance.model_time, 100.0 | units.Myr, 3)  # Unchanged!
        instance.stop()

    def test8(self):
        print("Testing adding and removing particles from stellar evolution code...")

        instance = MOBSE()
        instance.initialize_code()

        stars = Particles(6)
        stars.mass = [1.0, 1.0, 1.0, 0.2, 0.2, 0.2] | units.MSun

        binaries = Particles(3)
        binaries.eccentricity = 0.0
        for i in range(3):
            binaries[i].child1 = stars[i]
            binaries[i].child2 = stars[i+3]
        orbital_period = 200.0 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(
            orbital_period,
            binaries.child1.as_set().mass,
            binaries.child2.as_set().mass
        )
        binaries.semi_major_axis = semi_major_axis

        instance.commit_parameters()
        self.assertEqual(len(instance.particles), 0)
        self.assertEqual(len(instance.binaries), 0)  # before creation
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binaries[:-1])
        instance.commit_particles()
        instance.evolve_model(1.0 | units.Myr)
        self.assertEqual(len(instance.binaries), 2)  # before remove
        self.assertAlmostEqual(instance.binaries.age, 1.0 | units.Myr)

        instance.binaries.remove_particle(binaries[0])
        self.assertEqual(len(instance.binaries), 1)
        instance.evolve_model(2.0 | units.Myr)
        self.assertAlmostEqual(instance.binaries[0].age, 2.0 | units.Myr)

        instance.binaries.add_particles(binaries[::2])
        self.assertEqual(len(instance.binaries), 3)  # it's back...
        self.assertAlmostEqual(instance.binaries[0].age, 2.0 | units.Myr)
        self.assertAlmostEqual(instance.binaries[1].age, 0.0 | units.Myr)
        self.assertAlmostEqual(instance.binaries[2].age, 0.0 | units.Myr)  # ... and rejuvenated.

        instance.evolve_model(3.0 | units.Myr)  # The young stars keep their age offset from the old star
        self.assertAlmostEqual(instance.binaries.age, [3.0, 1.0, 1.0] | units.Myr)
        instance.evolve_model(4.0 | units.Myr)
        self.assertAlmostEqual(instance.binaries.age, [4.0, 2.0, 2.0] | units.Myr)
        instance.stop()

    def test9(self):
        print("Testing MOBSE states")
        instance = MOBSE()

        stars = Particles(2)
        stars.mass = [1.0, 0.2] | units.MSun

        binaries = Particles(1)
        orbital_period = 200.0 | units.day
        semi_major_axis = instance.orbital_period_to_semi_major_axis(orbital_period,  stars[0].mass, stars[1].mass)
        binaries.semi_major_axis = semi_major_axis
        binaries.eccentricity = 0.0
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        print("First do everything manually:", end=' ')
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()
        print("ok")

        print("initialize_code(), commit_parameters(), "
            "and cleanup_code() should be called automatically:", end=' ')
        instance = MOBSE()
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.reimers_mass_loss_coefficient = 0.5
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binaries)
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')
        print("ok")



import os
import sys
import time
from amuse.community.phigrape.interface import PhiGRAPEInterface, PhiGRAPE

from amuse.test.amusetest import TestWithMPI

import numpy
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.ic.plummer import new_plummer_model
try:
    import matplotlib
    matplotlib.use("Agg")
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False

default_test_options = dict()
# ~default_test_options = dict(mode='gpu')


class TestMPIInterface(TestWithMPI):

    def test0(self):
        instance = PhiGRAPEInterface(**default_test_options)
        self.assertTrue("Harfst, S., Gualandris, A., Merritt, D., Spurzem, R., Portegies Zwart, S., & Berczik, P."
            in instance.all_literature_references_string())
        instance.stop()

    def test1(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        instance.new_particle(11.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0)
        retrieved_state = instance.get_state(1)
        self.assertEqual(11.0,  retrieved_state['mass'])
        self.assertEqual(2.0, retrieved_state['radius'])
        self.assertEqual(instance.get_number_of_particles()['number_of_particles'], 1)
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        instance = PhiGRAPEInterface(**default_test_options)
        for x in [0.101, 4.0]:
            error = instance.set_eps2(x)
            self.assertEqual(error, 0)
            value, error = instance.get_eps2()
            self.assertEqual(error, 0)
            self.assertEqual(x, value)
        instance.cleanup_code()
        instance.stop()

    def test3(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()

        instance.new_particle([11.0, 12.0, 13.0, 14.0], [2.1, 3.1, 4.1, 5.1], [2.2, 3.2, 4.2, 5.2], [2.3, 3.3, 4.3, 5.3], [2.4, 3.4, 4.4, 5.4], [2.5, 3.5, 4.5, 5.5], [2.6, 3.6, 4.6, 5.6], [2.0, 3.0, 4.0, 5.0])
        retrieved_state = instance.get_state(1)
        self.assertEqual(11.0,  retrieved_state['mass'])
        retrieved_state = instance.get_state([2, 3, 4])
        self.assertEqual(12.0,  retrieved_state['mass'][0])
        self.assertEqual(instance.get_number_of_particles()['number_of_particles'], 4)
        instance.cleanup_code()
        instance.stop()

    def test5(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        n = 4000
        ids = [i for i in range(1, n)]
        values = [1.0 * i for i in range(1, n)]
        instance.new_particle(
              values, values, values, values, values, values, values)
        retrieved_state = instance.get_state(3999)
        self.assertEqual(3999.0,  retrieved_state['mass'])
        instance.cleanup_code()
        instance.stop()

    def test6(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        n = 4000
        ids = [i for i in range(1, n)]
        values = [1.0 * i for i in range(1, n)]
        for i in range(n-1):
            instance.new_particle(
                  values[i], values[i], values[i], values[i], values[i], values[i], values[i])

        retrieved_state = instance.get_state(1)
        self.assertEqual(1.0,  retrieved_state['mass'])
        instance.cleanup_code()
        instance.stop()

    def test7(self):
        instance = PhiGRAPEInterface(**default_test_options)  # (debugger="xterm")
        instance.initialize_code()

        instance.set_eps2(0.0**2)
        instance.set_eta(0.01, 0.02)
        instance.commit_parameters()

        instance.new_particle(
            [1.0, 1.0, 1.0],
            [1.0, 0.0, -1.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0])
        instance.recommit_particles()
        Ep = instance.get_potential_energy()['potential_energy']
        Ek = instance.get_kinetic_energy()['kinetic_energy']
        self.assertEqual(Ek, 0.5)
        self.assertEqual(Ep, -2.5)
        instance.delete_particle(2)
        instance.recommit_particles()
        n = instance.get_number_of_particles()['number_of_particles']
        Ep = instance.get_potential_energy()['potential_energy']
        Ek = instance.get_kinetic_energy()['kinetic_energy']

        self.assertEqual(n, 2)
        self.assertEqual(Ek, 0.)
        self.assertEqual(Ep, -0.5)

        instance.cleanup_code()
        instance.stop()

    def xtest8(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        instance.set_eps2(0.0**2)
        instance.set_eta(0.01, 0.02)
        instance.new_particle(
            [0.01, 0.01],
            [10., -10.],
            [0.0, 0.0],
            [0.0, 0.0],
            [-5.0, 5.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.1, 0.1])
        instance.commit_particles()
        # HAS NO RESULT...
        result = instance.evolve_model(3.14159)

        tnow = instance.get_time()['time']
        print("after evolve(pi), tnow = %f" % (tnow))
        # self.assertEqual( id1, 1)
        """
        instance.evolve(instance.get_time(),1)
        id2=instance.find_colliding_secondary(id1)
        self.assertEqual( id2, 2)
        self.assertAlmostEqual( tnow, 2.,2)
        state1 = instance.get_state(id1)
        state2 = instance.get_state(id2)
        self.assertTrue( abs(state1['x'] - state2['x'])<0.2)
        """
        instance.cleanup_code()
        instance.stop()

    def test8(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        instance.set_eps2(0.1**2)
        instance.commit_parameters()
        id1, errorcode = instance.new_particle(mass=10.0, radius=1.0,
            x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        id2, errorcode = instance.new_particle(mass=10.0, radius=1.0,
            x=2.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        potential, errorcode = instance.get_potential(id1)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential, -10.0 / numpy.sqrt(2.0**2 + 0.1**2), 6)
        total_potential, errorcode = instance.get_potential_energy()
        self.assertEqual(errorcode, 0)
        potentials, errorcode = instance.get_potential([id1, id2])
        instance.cleanup_code()
        instance.stop()
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(total_potential, numpy.sum(potentials * [10.0, 10.0]) / 2.0)

    def test9(self):
        instance = PhiGRAPEInterface(**default_test_options)
        instance.initialize_code()
        instance.set_eps2(0)
        instance.commit_parameters()
        id1, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        id2, errorcode = instance.new_particle(mass=1.0, radius=1.0, x=2.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        potential, errorcode = instance.get_potential(id1)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential,  -1.0 / numpy.sqrt(2.0**2), 8)
        total_potential, errorcode = instance.get_potential_energy()
        potentials, errorcode = instance.get_potential([id1, id2])
        instance.cleanup_code()
        instance.stop()

        self.assertAlmostRelativeEquals(total_potential, numpy.sum(potentials * [10.0, 1.0]) / 2.0)


class TestPhigrape(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = PhiGRAPE(convert_nbody, **default_test_options)  # , debugger="xterm")
        instance.initialize_code()
        instance.parameters.set_defaults()
        instance.parameters.initial_timestep_parameter = 0.001
        instance.parameters.timestep_parameter = 0.001

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)
        instance.commit_particles()

        position_at_start = earth.position.value_in(units.AU)[0]
        instance.evolve_model(365 | units.day)
        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 2)

        instance.evolve_model(365.0 + (365.0 / 2) | units.day)
        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 2)

        instance.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)
        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = PhiGRAPE(convert_nbody, **default_test_options)

        instance.initialize_code()
        instance.parameters.set_defaults()

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)
        instance.commit_particles()

        for x in range(1, 365, 30):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")

            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "phiGRAPE-earth-sun2.svg")
            figure.savefig(output_file)

        instance.cleanup_code()
        instance.stop()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = PhiGRAPE(convert_nbody, **default_test_options)

        instance.initialize_code()
        instance.parameters.set_defaults()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 2)

        instance.particles.mass = [17.0, 33.0] | units.kg

        self.assertEqual(instance.get_mass(1), 17.0 | units.kg)
        self.assertEqual(instance.get_mass(2), 33.0 | units.kg)

        instance.stop()

    def test5(self):
        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()

        particles = datamodel.Particles(2)
        particles.mass = [1.0, 1.0] | nbody_system.mass
        particles.radius = [0.0001, 0.0001] | nbody_system.length
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)

        instance.commit_particles()

        zero = 0.0 | nbody_system.length
        fx, fy, fz = instance.get_gravity_at_point(zero, 1.0 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(fx, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fy, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fz, 0.0 | nbody_system.acceleration, 3)
        for x in (0.25, 0.5, 0.75):
            x0 = x | nbody_system.length
            x1 = (2.0 - x) | nbody_system.length
            potential0 = instance.get_potential_at_point(zero, x0, zero, zero)
            potential1 = instance.get_potential_at_point(zero, x1, zero, zero)
            fx0, fy0, fz0 = instance.get_gravity_at_point(zero, x0, zero, zero)
            fx1, fy1, fz1 = instance.get_gravity_at_point(zero, x1, zero, zero)

            self.assertAlmostEqual(fy0, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fz0, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fy1, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fz1, 0.0 | nbody_system.acceleration, 3)

            self.assertAlmostEqual(fx0, -1.0 * fx1, 5)
            fx = (-1.0 / (x0**2) + 1.0 / (x1**2)) * (1.0 | nbody_system.length ** 3 / nbody_system.time ** 2)
            self.assertAlmostEqual(fx, fx0, 2)
            self.assertAlmostEqual(potential0, potential1, 5)

        instance.stop()

    def test6(self):
        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()

        particles = datamodel.Particles(6)
        particles.mass = nbody_system.mass.new_quantity(range(1, 7))
        particles.radius = 0.00001 | nbody_system.length
        particles.position = [[-1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0], [0.0, 0.0, 1.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)
        instance.commit_particles()
        copyof = instance.particles.copy()

        self.assertEqual(2 | nbody_system.mass, copyof[1].mass)

        instance.stop()

    def test7(self):
        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()

        particles = datamodel.Particles(2)
        particles.mass = [1.0, 1.0] | nbody_system.mass
        particles.radius = [0.0001, 0.0001] | nbody_system.length
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)

        instance.commit_particles()

        zero = [0.0, 0.0, 0.0] | nbody_system.length
        fx, fy, fz = instance.get_gravity_at_point(zero, [0.5, 1.0, 1.5] | nbody_system.length, zero, zero)
        self.assertAlmostRelativeEqual(fx[0], -3.55555555556 | nbody_system.acceleration, 5)
        self.assertAlmostRelativeEqual(fy[0], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fz[0], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fx[1], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fy[1], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fz[1], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fx[2], 3.55555555556 | nbody_system.acceleration, 5)
        self.assertAlmostRelativeEqual(fy[2], 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostRelativeEqual(fz[2], 0.0 | nbody_system.acceleration, 3)

        n = 512
        x = nbody_system.length.new_quantity(numpy.linspace(0.1, 1.9, n))
        zero = nbody_system.length.new_quantity(numpy.zeros(n))
        fx, fy, fz = instance.get_gravity_at_point(zero, x, zero, zero)
        for i in range(n//2):
            self.assertAlmostRelativeEqual(fx[i], - fx[n - 1 - i], 5)

        instance.stop()

    def test8(self):
        particles = datamodel.Particles(6)
        particles.mass = nbody_system.mass.new_quantity(range(1, 7))
        particles.radius = 0.00001 | nbody_system.length
        particles.position = [[-1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0], [0.0, 0.0, 1.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed

        for current_mode in ['g6lib', 'gpu', 'grape']:
            try:
                instance = PhiGRAPE(mode=current_mode)
            except:
                print("Running PhiGRAPE with mode=" + current_mode, "was unsuccessful.")
            else:
                print("Running PhiGRAPE with mode=" + current_mode + "...", end=' ')
                instance.initialize_code()
                instance.particles.add_particles(particles)
                instance.commit_particles()
                instance.evolve_model(0.1 | nbody_system.time)
                instance.cleanup_code()
                instance.stop()
                print("ok")

    def test9(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = PhiGRAPE(convert_nbody, **default_test_options)  # , debugger="xterm")
        instance.initialize_code()

        instance.parameters.set_defaults()
        instance.parameters.initialize_gpu_once = 1

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        instance.particles.add_particles(stars)
        instance.commit_particles()

        instance.evolve_model(365 | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)

        position_at_start = earth.position.value_in(units.AU)[0]
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 6)

        instance.evolve_model(365.0 + (365.0 / 2) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 2)

        instance.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        # self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        instance.cleanup_code()

        instance.stop()

    def test10(self):
        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()

        stars = new_plummer_model(100)
        stars.radius = 0 | nbody_system.length

        instance.particles.add_particles(stars)
        channel = stars.new_channel_to(instance.particles)

        instance.evolve_model(0.001 | nbody_system.time)

        e0 = instance.kinetic_energy + instance.potential_energy

        stars.mass *= 0.9
        channel.copy()

        instance.synchronize_model()

        e1 = instance.kinetic_energy + instance.potential_energy

        delta_e = e1 - e0

        self.assertTrue(e1 != e0)

        instance.stop()

    def test11(self):
        print("Testing PhiGRAPE collision_detection")
        particles = datamodel.Particles(7)
        particles.mass = 0.001 | nbody_system.mass
        particles.radius = 0.01 | nbody_system.length
        particles.x = [-101.0, -100.0, -0.5, 0.5, 100.0, 101.0, 104.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [[2, 0, 0], [-2, 0, 0]]*3 + [[-4, 0, 0]] | nbody_system.speed

        instance = PhiGRAPE(redirection='none', **default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()
        instance.particles.add_particles(particles)
        collisions = instance.stopping_conditions.collision_detection
        collisions.enable()
        instance.evolve_model(1.0 | nbody_system.time)

        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 0.5 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 3)
        self.assertEqual(len(collisions.particles(1)), 3)
        self.assertEqual(len(particles - collisions.particles(0) - collisions.particles(1)), 1)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True, True, True])

        sticky_merged = datamodel.Particles(len(collisions.particles(0)))
        sticky_merged.mass = collisions.particles(0).mass + collisions.particles(1).mass
        sticky_merged.radius = collisions.particles(0).radius
        for p1, p2, merged in zip(collisions.particles(0), collisions.particles(1), sticky_merged):
            merged.position = (p1 + p2).center_of_mass()
            merged.velocity = (p1 + p2).center_of_mass_velocity()

        print(instance.model_time)
        print(instance.particles)
        instance.particles.remove_particles(collisions.particles(0) + collisions.particles(1))
        instance.particles.add_particles(sticky_merged)

        instance.evolve_model(1.0 | nbody_system.time)
        print()
        print(instance.model_time)
        print(instance.particles)
        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 1.0 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 1)
        self.assertEqual(len(collisions.particles(1)), 1)
        self.assertEqual(len(instance.particles - collisions.particles(0) - collisions.particles(1)), 2)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True])
        instance.stop()

    def test13(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.stopping_conditions_number_of_steps = 2
        self.assertEqual(instance.parameters.stopping_conditions_number_of_steps, 2)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(10 | nbody_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        self.assertTrue(instance.model_time < 10 | nbody_system.time)

        instance.stop()

    def test14(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        very_short_time_to_evolve = 1 | units.s
        very_long_time_to_evolve = 1e9 | nbody_system.time

        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.stopping_conditions_timeout = very_short_time_to_evolve
        self.assertEqual(instance.parameters.stopping_conditions_timeout, very_short_time_to_evolve)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.timeout_detection.enable()
        start = time.time()
        instance.evolve_model(very_long_time_to_evolve)
        end = time.time()
        self.assertTrue(instance.stopping_conditions.timeout_detection.is_set())
        self.assertTrue((end-start) < very_short_time_to_evolve.value_in(units.s) + 2)  # 2 = some overhead compensation

        instance.stop()

    def test15(self):
        instance = PhiGRAPE(number_of_workers=2, **default_test_options)  # , redirection = "none")
        instance.initialize_code()
        instance.parameters.set_defaults()
        print(1)
        particles = datamodel.Particles(2)
        particles.x = [-1.0, 1.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance.particles.add_particles(particles)
        instance.commit_particles()

        instance.evolve_model(0.01 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertEqual(instance.particles[0].position.x, -instance.particles[1].position.x)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x, 0.999969482111 | nbody_system.length, 6)

        instance.evolve_model(0.10 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertEqual(instance.particles[0].position.x, -instance.particles[1].position.x)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x, 0.99804560161 | nbody_system.length, 6)

        instance.evolve_model(0.50 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertEqual(instance.particles[0].position.x, -instance.particles[1].position.x)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x, 0.968416814302 | nbody_system.length, 6)

        instance.cleanup_code()

        instance.stop()

    def test16(self):
        instance = PhiGRAPE(number_of_workers=2, **default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()
        print(1)
        particles = datamodel.Particles(2)
        particles.x = [-1.0, 1.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = [1.0, 0.5] | nbody_system.mass

        instance.particles.add_particles(particles)
        instance.commit_particles()

        instance.evolve_model(0.01 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertAlmostRelativeEquals(instance.particles[0].position.x, -0.999984741095 | nbody_system.length, 8)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x,  0.999969482189 | nbody_system.length, 8)

        instance.evolve_model(0.10 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertAlmostRelativeEquals(instance.particles[0].position.x, -0.999022960148 | nbody_system.length, 8)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x, 0.998045920305 | nbody_system.length, 8)

        instance.evolve_model(0.50 | nbody_system.time)

        instance.particles.copy_values_of_all_attributes_to(particles)

        print(instance.particles.position.x)
        self.assertAlmostRelativeEquals(instance.particles[0].position.x, -0.984250788742 | nbody_system.length, 8)
        self.assertAlmostRelativeEquals(instance.particles[1].position.x, 0.968501583383 | nbody_system.length, 8)

        instance.cleanup_code()

        instance.stop()

    def test17(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)
        instance = PhiGRAPE(convert_nbody, **default_test_options)
        instance.initialize_code()

        instance.parameters.set_defaults()
        instance.parameters.initial_timestep_parameter = 0.001
        instance.parameters.timestep_parameter = 0.001
        instance.parameters.epsilon_squared = 0.0 | units.AU**2

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        instance.particles.add_particles(stars)
        instance.commit_particles()

        position_at_start = earth.position.value_in(units.AU)[0]
        print(instance.particles[0].position.as_quantity_in(units.AU))
        print(instance.particles[1].position.as_quantity_in(units.AU))
        instance.evolve_model(365 | units.day)
        print(instance.particles[0].position.as_quantity_in(units.AU))
        print(instance.particles[1].position.as_quantity_in(units.AU))

        instance.particles.copy_values_of_all_attributes_to(stars)

        position_after_full_rotation = earth.position.value_in(units.AU)[0]

        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 3)

        print(instance.particles[0].position.as_quantity_in(units.AU))
        instance.evolve_model(365.0 + (365.0 / 2) | units.day)

        print(instance.particles[0].position.as_quantity_in(units.AU))

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 2)

        instance.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        # self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        instance.cleanup_code()

        instance.stop()

    def test18(self):
        print("Testing PhiGRAPE delete_particle")
        number_of_particles = 15
        delete_order = [6, 9, 3, 1, 0, 2, 5]
        number_of_initial_deletes = 3

        particles = datamodel.Particles(number_of_particles)
        particles.mass = range(1, number_of_particles + 1) | nbody_system.mass
        particles.radius = 0.01 | nbody_system.length
        particles.x = range(number_of_particles) | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.vx = range(number_of_particles) | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed

        instance = PhiGRAPE(**default_test_options)
        instance.initialize_code()
        instance.parameters.set_defaults()
        initial_number_of_particles = number_of_particles - len(delete_order) + number_of_initial_deletes
        instance.particles.add_particles(particles[:initial_number_of_particles])
        index_new_particle = initial_number_of_particles

        self.assertEqual(len(instance.particles), initial_number_of_particles)
        instance.evolve_model(0.01 | nbody_system.time)
        instance.particles.remove_particles(particles[delete_order[:number_of_initial_deletes]])
        number_of_deletes = number_of_initial_deletes
        self.assertEqual(len(instance.particles), initial_number_of_particles - number_of_deletes)

        for i, particle_index in enumerate(delete_order[number_of_initial_deletes:]):
            instance.evolve_model((i+1) / 10.0 | nbody_system.time)
            instance.particles.remove_particle(particles[particle_index])
            number_of_deletes += 1
            instance.particles.add_particle(particles[index_new_particle])
            index_new_particle += 1
            self.assertEqual(len(instance.particles), initial_number_of_particles - number_of_deletes + i + 1)
            self.assertAlmostEqual(particles[:index_new_particle].total_mass() - instance.particles.total_mass(),
                (delete_order[:number_of_deletes] | nbody_system.mass).sum() + (number_of_deletes | nbody_system.mass))

        instance.stop()

    def test19(self):
        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.parsec)

        particles = datamodel.Particles(2)
        particles.mass = 100 | units.MSun
        particles.radius = 200 | units.RSun
        particles[0].position = [0, 0, 0] | units.parsec
        particles[1].position = [1, 0, 0] | units.parsec
        particles.velocity = [0, 0, 0] | units.km / units.s

        code = PhiGRAPE(
            converter,
            PhiGRAPEInterface.MODE_G6LIB,
            number_of_workers=2
        )
        code.initialize_code()
        stop_cond = code.stopping_conditions.collision_detection
        stop_cond.enable()

        code.particles.add_particles(particles)

        code.evolve_model(0.08 | nbody_system.time)

        self.assertTrue(stop_cond.is_set())



from amuse.test import amusetest
from amuse.datamodel.incode_storage import *

import numpy
import time

from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system


class TestParticles(amusetest.TestCase):

    def test1(self):
        class Code(object):
            def __init__(self):
                # x,y,z,mass
                self.data = []
                self.get_position_called = False
                self.set_position_called = False

            def get_number_of_particles(self):
                return 0 if not self.data else len(self.data[0])

            def get_position(self, index):
                self.get_position_called = True
                data_to_return = [(self.data[0][i], self.data[1][i], self.data[2][i]) for i in index]
                data_to_return = numpy.asarray(data_to_return).reshape(3, -1)
                return [units.m(x) for x in data_to_return]

            def set_position(self, index, x, y, z):
                self.set_position_called = True
                pass

            def new_particle(self, x, y, z):
                x = x.value_in(units.m)
                y = y.value_in(units.m)
                z = z.value_in(units.m)
                self.data = [x, y, z]
                return [i for i in range(len(x))]

        code = Code()
        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("x", "y", "z")),
            None,
            code.get_number_of_particles,
            [],
            [ParticleGetAttributesMethod(code.get_position, ("x", "y", "z")),],
            name_of_the_index="index"
        )

        self.assertEqual(len(storage), 0)
        self.assertEqual(storage.get_defined_attribute_names(), ["x", "y", "z"])

        self.assertFalse(code.get_position_called)
        storage.get_values_in_store([], ["x", "y", "z"])
        self.assertFalse(code.get_position_called)

        storage.add_particles_to_store(
            [1, 2, 3, 4],
            ["x", "y", "z"],
            [
                units.m([1, 2, 3, 4]),
                units.m([2, 3, 4, 5]),
                units.m([3, 4, 5, 6])
            ]
        )

        self.assertEqual(len(storage), 4)

    def test2(self):
        class Code(object):
            def __init__(self):
                # x,y,z,mass
                self.data = []
                self.get_position_called = False
                self.set_position_called = False
                self.get_mass_called = False
                self.set_mass_called = False

            def get_number_of_particles(self):
                return 0 if not self.data else len(self.data[0])

            def get_position(self, index):
                self.get_position_called = True
                data_to_return = [(self.data[0][i], self.data[1][i], self.data[2][i]) for i in index]
                data_to_return = numpy.asarray(data_to_return).reshape(3, -1)
                return [units.m(x) for x in data_to_return]

            def get_mass(self, index):
                self.get_mass_called = True
                data_to_return = [self.data[3][i] for i in index]
                return units.kg(data_to_return)

            def set_position(self, index, x, y, z):
                self.set_position_called = True
                pass

            def set_mass(self, index, mass):
                self.set_mass_called = True
                pass

            def new_particle(self, x, y, z, mass):
                x = x.value_in(units.m)
                y = y.value_in(units.m)
                z = z.value_in(units.m)
                mass = mass.value_in(units.kg)
                self.data = [x, y, z, mass]
                return [i for i in range(len(x))]

        code = Code()
        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("x", "y", "z", "mass")),
            None,
            code.get_number_of_particles,
            [],
            [
                ParticleGetAttributesMethod(code.get_position, ("x", "y", "z")),
                ParticleGetAttributesMethod(code.get_mass, ("mass",)),
            ],
            name_of_the_index="index"
        )

        storage.add_particles_to_store(
            [1, 2, 3, 4],
            ["x", "y", "z", "mass"],
            [
                units.m([1, 2, 3, 4]),
                units.m([2, 3, 4, 5]),
                units.m([3, 4, 5, 6]),
                units.kg([13, 14, 15, 16]),
            ]
        )

        self.assertEqual(len(storage), 4)

        self.assertEqual(storage.get_defined_attribute_names(), ["mass", "x", "y", "z"])

        self.assertFalse(code.get_position_called)
        self.assertFalse(code.get_mass_called)
        indices = storage.get_indices_of([2, 3])
        x, y, mass = storage.get_values_in_store(indices, ["x", "y", "mass"])
        self.assertTrue(code.get_position_called)
        self.assertTrue(code.get_mass_called)
        self.assertEqual(x[1], 3 | units.m)
        self.assertEqual(mass[1], 15 | units.kg)

    def test3(self):
        class Code(object):
            def __init__(self):
                # mass
                self.data = []
                self.get_mass_called = False
                self.set_mass_called = False

            def get_number_of_particles(self):
                return 0 if not self.data else len(self.data[0])

            def get_mass(self, index):
                self.get_mass_called = True
                data_to_return = [self.data[0][i] for i in index]
                return units.kg(data_to_return)

            def set_mass(self, index, mass):
                self.set_mass_called = True
                pass

            def new_particle(self, mass):
                mass = mass.value_in(units.kg)
                self.data = [mass]
                return [i for i in range(len(mass))]

        code = Code()
        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("mass",)),
            None,
            code.get_number_of_particles,
            [],
            [
                ParticleGetAttributesMethod(code.get_mass, ("mass",)),
            ],
            name_of_the_index="index"
        )

        storage.add_particles_to_store(
            [1, 2, 3, 4],
            ["mass"],
            [
                units.kg([1, 2, 3, 4]),
            ]
        )

        self.assertEqual(len(storage), 4)

        self.assertEqual(storage.get_defined_attribute_names(), ["mass",])

        indices = storage.get_indices_of([2, 3])
        index, mass = storage.get_values_in_store(indices, ["index_in_code", "mass"])
        self.assertTrue(code.get_mass_called)
        self.assertEqual(index[0], 1)
        self.assertEqual(mass[0],  2 | units.kg)
        self.assertEqual(index[1], 2)
        self.assertEqual(mass[1],  3 | units.kg)

    def test4(self):
        class Code(object):
            def __init__(self):
                # mass
                self.data = []
                self.get_mass_called = False
                self.set_mass_called = False
                self.number_of_particles = 0

            def get_number_of_particles(self):
                return self.number_of_particles

            def get_mass(self, index):
                self.get_mass_called = True
                data_to_return = [self.data[i] for i in index]
                return units.kg(data_to_return)

            def set_mass(self, index, mass):
                self.set_mass_called = True
                pass

            def new_particle(self, mass):
                mass = mass.value_in(units.kg)
                self.data = mass
                self.number_of_particles = len(self.data)
                return [i for i in range(len(mass))]

        code = Code()
        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("mass",)),
            None,
            code.get_number_of_particles,
            [],
            [
                ParticleGetAttributesMethod(code.get_mass, ("mass",)),
            ],
            name_of_the_index="index"
        )

        storage.add_particles_to_store(
            numpy.asarray([1, 2, 3, 4], dtype='uint64'),
            ["mass"],
            [
                units.kg([1, 2, 3, 4]),
            ]
        )

        self.assertEqual(len(storage), 4)

        storage._remove_indices([1, 2,])
        code.number_of_particles = 2
        indices = storage.get_indices_of([1, 4])
        index, mass = storage.get_values_in_store(indices, ["index_in_code", "mass"])

        self.assertEqual(index[0], 0)
        self.assertEqual(index[1], 3)
        self.assertEqual(mass[0],  1 | units.kg)
        self.assertEqual(mass[1],  4 | units.kg)

        self.assertEqual(len(storage), 2)

        storage._add_indices([4, 5])
        code.data = numpy.concatenate((code.data, [5, 6]))

        code.number_of_particles = 4
        self.assertEqual(len(storage), 4)

        indices = storage.get_indices_of(storage.particle_keys)
        mass, = storage.get_values_in_store(indices, ["mass"])

        self.assertEqual(mass[0],  1 | units.kg)
        self.assertEqual(mass[1],  4 | units.kg)
        self.assertEqual(mass[2],  5 | units.kg)
        self.assertEqual(mass[3],  6 | units.kg)

        storage._remove_indices([4,])
        code.number_of_particles = 3
        self.assertEqual(len(storage), 3)

        indices = storage.get_indices_of(storage.particle_keys)
        mass, = storage.get_values_in_store(indices, ["mass"])

        self.assertEqual(mass[0],  1 | units.kg)
        self.assertEqual(mass[1],  4 | units.kg)
        self.assertEqual(mass[2],  6 | units.kg)

    def test5(self):
        class Code(object):
            def __init__(self):
                self.data = []
                self.number_of_particles = 0

            def get_number_of_particles(self):
                return self.number_of_particles

            def get_mass(self, index):
                data_to_return = [self.data[i][0] for i in index]
                return units.kg(data_to_return)

            def get_children(self, index):
                return [(self.data[i][1]) for i in index], [(self.data[i][2]) for i in index]

            def new_particle(self, mass):
                mass = mass.value_in(units.kg)
                self.data = [[x, -1, -1] for x in mass]
                self.number_of_particles = len(self.data)
                return [i for i in range(len(mass))]

        code = Code()

        children_getter = ParticleGetAttributesMethod(
                    code.get_children,
                    ('child1', 'child2',)
        )
        children_getter.index_output_attributes = set(['child1', 'child2'])

        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("mass",)),
            None,
            code.get_number_of_particles,
            [],
            [
                ParticleGetAttributesMethod(code.get_mass, ("mass",)),
                children_getter
            ],
            name_of_the_index="index"
        )

        storage.add_particles_to_store(
            numpy.asarray([100, 200, 300, 400], dtype='uint64'),
            ["mass"],
            [
                units.kg([1, 2, 3, 4]),
            ]
        )

        self.assertEqual(len(storage), 4)

        indices = storage.get_indices_of([100, 400])
        mass = storage.get_values_in_store(indices, ["mass",])[0]
        self.assertEqual(mass[0], 1.0 | units.kg)
        self.assertEqual(mass[1], 4.0 | units.kg)

        code.data[0][1] = 1
        code.data[0][2] = 2

        indices = storage.get_indices_of([100])
        child1, child2 = storage.get_values_in_store(indices, ['child1', 'child2'])

        self.assertEqual(child1[0].number, 200)
        self.assertEqual(child2[0].number, 300)

    def test7(self):
        class Code(object):
            def __init__(self):
                # x,y,z,mass
                self.data = []
                self.get_position_called = False
                self.set_position_called = False
                self.get_mass_called = False
                self.set_mass_called = False

            def get_number_of_particles(self):
                return 0 if not self.data else len(self.data[0])

            def get_position(self, index):
                self.get_position_called = True
                data_to_return = [(self.data[0][i], self.data[1][i], self.data[2][i]) for i in index]
                data_to_return = numpy.asarray(data_to_return).reshape(3, -1)
                return [units.m(x) for x in data_to_return]

            def get_mass(self, index):
                self.get_mass_called = True
                data_to_return = [self.data[3][i] for i in index]
                return data_to_return

            def set_position(self, index, x, y, z):
                self.set_position_called = True
                pass

            def set_mass(self, index, mass):
                self.set_mass_called = True
                for i, j in enumerate(index):
                    self.data[3][j] = mass[i]
                return [0 for i in range(len(index))]

            def new_particle(self, x, y, z, mass):
                x = x.value_in(units.m)
                y = y.value_in(units.m)
                z = z.value_in(units.m)
                mass = mass
                self.data = [x, y, z, mass]
                return [i for i in range(len(x))]

        code = Code()
        storage = InCodeAttributeStorage(
            code,
            NewParticleMethod(code.new_particle, ("x", "y", "z", "mass")),
            None,
            code.get_number_of_particles,
            [
                ParticleSetAttributesMethod(code.set_position, ("x", "y", "z")),
                ParticleSetAttributesMethod(code.set_mass, ("mass",)),
            ],
            [
                ParticleGetAttributesMethod(code.get_position, ("x", "y", "z")),
                ParticleGetAttributesMethod(code.get_mass, ("mass",)),
            ],
            name_of_the_index="index"
        )

        storage.add_particles_to_store(
            [1, 2, 3, 4],
            ["x", "y", "z", "mass"],
            [
                units.m([1, 2, 3, 4]),
                units.m([2, 3, 4, 5]),
                units.m([3, 4, 5, 6]),
                numpy.asarray([13.0, 14.0, 15, 16]),
            ]
        )

        self.assertEqual(len(storage), 4)

        self.assertEqual(storage.get_defined_attribute_names(), ["mass", "x", "y", "z"])

        self.assertFalse(code.get_position_called)
        self.assertFalse(code.get_mass_called)
        indices = storage.get_indices_of([2, 3])
        x, y, mass = storage.get_values_in_store(indices, ["x", "y", "mass"])
        self.assertTrue(code.get_position_called)
        self.assertTrue(code.get_mass_called)
        self.assertEqual(x[1], 3 | units.m)
        self.assertEqual(mass[1], 15)
        self.assertEqual(mass[0], 14)
        storage.set_values_in_store(indices, ["x", "y", "z", "mass"], [[10, 11] | units.m, [12, 14] | units.m, [12, 14] | units.m, [40.0, 50.0]])
        x, y, mass = storage.get_values_in_store(indices, ["x", "y", "mass"])
        self.assertEqual(mass[1], 50)
        self.assertEqual(mass[0], 40)


class TestGrids(amusetest.TestCase):

    def test1(self):
        class Code(object):
            def get_range(self):
                return (1, 10, 2, 5, 3, 6)

            def get_ijk(self, i, j, k):
                return units.m(i), units.m(j), units.m(k)

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [],
            [ParticleGetAttributesMethod(code.get_ijk, ("i", "j", "k")),],
        )

        self.assertEqual(storage.storage_shape(), (10, 4, 4))
        self.assertEqual(storage.get_defined_attribute_names(), ["i", "j", "k"])

        values = storage.get_values_in_store((0, 1, 1), ("i",))
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0], 1 | units.m)

        values = storage.get_values_in_store((0, 1, 1), ("k", "j", "i",))
        self.assertEqual(values[0], 4 | units.m)
        self.assertEqual(values[1], 3 | units.m)
        self.assertEqual(values[2], 1 | units.m)

    def test2(self):
        class Code(object):
            def get_range(self):
                return (1, 10, 2, 5, 3, 6)

            def get_ijk(self, i, j, k):
                return units.m(i), units.m(j), units.m(k)

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [],
            [ParticleGetAttributesMethod(code.get_ijk, ("i", "j", "k")),],
        )
        values = storage.get_values_in_store(numpy.s_[0:2], ("i",))
        self.assertEqual(len(values), 1)
        self.assertEqual(len(values[0]), 2)
        self.assertEqual(values[0].number.shape, (2, 4, 4))
        self.assertEqual(values[0][0][0][0], 1 | units.m)
        self.assertEqual(values[0][1][0][0], 2 | units.m)

    def test3(self):

        shape = (11, 5, 5)

        class Code(object):

            def __init__(self):
                self.storage = numpy.arange(shape[0]*shape[1]*shape[2]).reshape(shape)

            def get_range(self):
                return (0, shape[0]-1, 0, shape[1]-1, 0, shape[2]-1)

            def get_a(self, i_s, j_s, k_s):
                return units.m.new_quantity(numpy.asarray([(self.storage[i][j][k]) for i, j, k in zip(i_s, j_s, k_s)]))

            def set_a(self, i_s, j_s, k_s, values):
                # ~ print i_s, j_s, k_s
                # ~ print "VALUES:", values
                index = 0
                for i, j, k in zip(i_s, j_s, k_s):
                    self.storage[i][j][k] = values[index].value_in(units.m)
                    index += 1
                    # ~ print index

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [ParticleSetAttributesMethod(code.set_a, ("a",)),],
            [ParticleGetAttributesMethod(code.get_a, ("a",)),],
        )

        values = storage.get_values_in_store(None, ("a",))
        self.assertTrue(numpy.all(values[0].value_in(units.m) == code.storage))
        # self.assertTrue(False)
        values = storage.get_values_in_store((0, 0, 0), ("a",))
        self.assertEqual(values[0], 0 | units.m)
        storage.set_values_in_store((0, 0, 0), ("a",), [11.0 | units.m,])
        values = storage.get_values_in_store((0, 0, 0), ("a",))
        self.assertEqual(values[0], 11.0 | units.m)
        values = storage.get_values_in_store((0, 0), ("a",))
        storage.set_values_in_store((0, 0), ("a",), [[11.0, 12.0, 13.0, 14.0, 15.0] | units.m,])
        self.assertTrue(numpy.all(code.storage[0][0] == [11.0, 12.0, 13.0, 14.0, 15.0]))

    def test4(self):
        class Code(object):
            def get_range(self, d, l):
                return (1, 10, 2, 5, 3, 6)

            def get_ijk(self, i, j, k, d, l):
                return units.m(d), units.m(l), units.m(k)

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [],
            [ParticleGetAttributesMethod(code.get_ijk, ("i", "j", "k")),],
            extra_keyword_arguments_for_getters_and_setters={'d': 1, 'l': 2},
        )

        self.assertEqual(storage.storage_shape(), (10, 4, 4))
        self.assertEqual(storage.get_defined_attribute_names(), ["i", "j", "k"])

        values = storage.get_values_in_store((0, 1, 1), ("i",))
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0], 1 | units.m)

        values = storage.get_values_in_store((0, 1, 1), ("k", "j", "i",))
        self.assertEqual(values[0], 4 | units.m)
        self.assertEqual(values[1], 2 | units.m)
        self.assertEqual(values[2], 1 | units.m)

    def test5(self):
        class Code(object):
            def get_range(self):
                return (1, 10, 2, 5, 3, 6)

            def get_ijk(self, i, j, k):
                return units.m(i), units.m(j), units.m(k)

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [],
            [ParticleGetAttributesMethod(code.get_ijk, ("i", "j", "k")),],
        )

        self.assertEqual(storage.storage_shape(), (10, 4, 4))
        self.assertEqual(storage.get_defined_attribute_names(), ["i", "j", "k"])

        values = storage.get_values_in_store(None, ("i",))
        self.assertEqual(len(values), 1)
        self.assertEqual(values[0].number.ndim, 3)

    def test6(self):

        shape = (11, 5, 5)

        class Code(object):

            def __init__(self):
                self.storage = numpy.arange(shape[0]*shape[1]*shape[2]).reshape(shape)

            def get_range(self):
                return (0, shape[0]-1, 0, shape[1]-1, 0, shape[2]-1)

            def get_a(self, i_s, j_s, k_s):
                return numpy.asarray([(self.storage[i][j][k]) for i, j, k in zip(i_s, j_s, k_s)])

            def set_a(self, i_s, j_s, k_s, values):
                # ~ print i_s, j_s, k_s
                # ~ print "VALUES:", values
                index = 0
                for i, j, k in zip(i_s, j_s, k_s):
                    self.storage[i][j][k] = values[index]
                    index += 1
                    # ~ print index

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [ParticleSetAttributesMethod(code.set_a, ("a",)),],
            [ParticleGetAttributesMethod(code.get_a, ("a",)),],
        )

        values = storage.get_values_in_store(None, ("a",))
        self.assertTrue(numpy.all(values[0] == code.storage))
        values = storage.get_values_in_store((0, 0, 0), ("a",))
        self.assertEqual(values[0], 0)
        storage.set_values_in_store((0, 0, 0), ("a",), [11.0,])
        values = storage.get_values_in_store((0, 0, 0), ("a",))
        self.assertEqual(values[0], 11.0)
        values = storage.get_values_in_store((0, 0), ("a",))[0]
        self.assertTrue(numpy.all(values == [11.0, 1.0, 2.0, 3.0, 4.0]))
        storage.set_values_in_store((0, 0), ("a",), [[11.0, 12.0, 13.0, 14.0, 15.0],])
        self.assertTrue(numpy.all(code.storage[0][0] == [11.0, 12.0, 13.0, 14.0, 15.0]))

    def test7(self):

        shape = (11, 5, 5)

        class Code(object):

            def __init__(self):
                self.storage = numpy.arange(shape[0]*shape[1]*shape[2]).reshape(shape)

            def get_range(self):
                return (0, shape[0]-1, 0, shape[1]-1, 0, shape[2]-1)

            def get_a(self, i_s, j_s, k_s):
                return numpy.asarray([(self.storage[i][j][k]) for i, j, k in zip(i_s, j_s, k_s)])

            def set_a(self, i_s, j_s, k_s, values):
                index = 0
                for i, j, k in zip(i_s, j_s, k_s):
                    self.storage[i][j][k] = values[index]
                    index += 1

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [ParticleSetAttributesMethod(code.set_a, ("a",)),],
            [ParticleGetAttributesMethod(code.get_a, ("a",)),],
        )

        values = storage.get_values_in_store((), ())
        self.assertTrue(values == [])
        values = storage.get_values_in_store((0, 0, 1,), ("a",))
        self.assertTrue(values[0] == 1)

    def test8(self):
        class Code(object):
            def __init__(self):
                self.storage = 1. | units.m

            def get_range(self):
                return ()

            def get_a(self):
                return self.storage

            def set_a(self, value):
                self.storage = value

        code = Code()

        storage = InCodeGridAttributeStorage(
            code,
            code.get_range,
            [ParticleSetAttributesMethod(code.set_a, ("a",)),],
            [ParticleGetAttributesMethod(code.get_a, ("a",)),],
        )

        self.assertEqual(storage.storage_shape(), ())
        self.assertEqual(storage.get_defined_attribute_names(), ['a'])

        values = storage.get_values_in_store((), ("a",))
        self.assertEqual(len(values), 1)
        print(values, "<")
        self.assertEqual(values[0], 1 | units.m)



from amuse.test.amusetest import TestCase
from amuse.support.exceptions import AmuseException
from amuse.units import units
import sys


class TestAmusetestTestCase(TestCase):

    def test1(self):
        self.assertEqual(1 | units.m, 1 | units.m)
        self.assertEqual(1 | units.m, 1.0 | units.m)
        self.assertEqual([1, 2, 3] | units.m, [1, 2, 3] | units.m)
        self.assertEqual([1, 1, 1] | units.m, 1 | units.m)
        self.assertEqual(1 | units.m, [1, 1, 1] | units.m)
        self.assertRaises(self.failureException, self.assertEqual, 1 | units.m, 0 | units.m)
        self.assertRaises(self.failureException, self.assertEqual,
            1 | units.m, 0 | units.m, expected_message="1 m != 0 m")

    def test2(self):
        self.assertAlmostEqual(1 | units.m, 1 | units.m)
        self.assertAlmostEqual(1 | units.m, 1.000000049 | units.m)
        self.assertAlmostEqual([1, 2, 3] | units.m, [1, 2, 3] | units.m)
        self.assertAlmostEqual([1, 1, 1] | units.m, 1.000000049 | units.m)
        self.assertAlmostEqual(1 | units.m, [1, 1, 1.000000049] | units.m)
        self.assertAlmostEqual(1 | units.m, 1.00049 | units.m, places=3)
        self.assertRaises(self.failureException, self.assertAlmostEqual, 1.0 | units.m,
            1.000000051 | units.m, expected_message="1.0 m != 1.000000051 m within 7 places")
        self.assertRaises(self.failureException, self.assertAlmostEqual,
            [1.1, 1.0, 0.9] | units.m, 1.0 | units.m,
            expected_message="@0, 1.1 m != 1.0 m within 7 places\n@2, 0.9 m != 1.0 m within 7 places")

    def test3(self):
        self.assertAlmostRelativeEqual(1 | units.m, 1 | units.m)
        self.assertAlmostRelativeEqual(1 | units.m, 1.0 + 4.9e-16 | units.m)
        self.assertAlmostRelativeEqual([1, 2, 3] | units.m, [1, 2, 3] | units.m)
        self.assertAlmostRelativeEqual([1, 1, 1] | units.m, 1 | units.m)
        self.assertAlmostRelativeEqual(1 | units.m, [1, 1, 1] | units.m)
        self.assertAlmostRelativeEqual(100.0 | units.m, 100.1 | units.m, places=3)
        self.assertRaises(self.failureException, self.assertAlmostRelativeEqual,
            1.0 | units.m, [0.0, 1.0001, 1.0011] | units.m, places=3, expected_message="@0, quantity<1.0 m> != quantity<0.0 m> within 3 places\n"
            "@2, quantity<1.0 m> != quantity<1.0011 m> within 3 places")

    def test4(self):
        self.assertIsOfOrder(1 | units.m, 1 | units.m)
        self.assertIsOfOrder(1 | units.m, 3.0 | units.m)
        self.assertIsOfOrder([1, 2, 3] | units.m, [3, 2, 1] | units.m)
        self.assertIsOfOrder([1, 1, 1] | units.m, 1 | units.m)
        self.assertIsOfOrder(1 | units.m, [0.4, 1, 3] | units.m)
        self.assertRaises(self.failureException, self.assertIsOfOrder,
            1.0 | units.m, [0.3, 1.0, 4.0] | units.m, expected_message="@0, 1.0 m is not of order 0.3 m\n@2, 1.0 m is not of order 4.0 m")

    def test5(self):
        self.assertRaises(AmuseException, (1 | units.m).as_quantity_in, units.kg)
        self.assertRaises(AmuseException, (1 | units.m).as_quantity_in, 1 | units.cm)
        self.assertRaises(AmuseException, (1 | units.m).as_quantity_in, 1 | units.cm,
            expected_message="Cannot expres a unit in a quantity")
        # check that a failureException is raised, when we erroneously assert an exception:

        if sys.hexversion > 0x03000000:
            error_message = "AmuseException not raised by as_quantity_in"
        else:
            error_message = "AmuseException not raised"
        self.assertRaises(self.failureException, self.assertRaises, AmuseException,
            (1 | units.m).as_quantity_in, units.cm, expected_message=error_message)
        # the ultimate test... lol

        if sys.hexversion > 0x03000000:
            error_message = "AssertionError not raised by assertRaises"
        else:
            error_message = "AssertionError not raised"

        self.assertRaises(self.failureException, self.assertRaises, self.failureException,
            self.assertRaises, AmuseException, (1 | units.m).as_quantity_in,
            1 | units.cm, expected_message=error_message)
        self.assertRaises(AmuseException, lambda: 1 + (1 | units.m))
        self.assertRaises(ZeroDivisionError, lambda: 1 | units.m/0)

    def test6(self):
        self.assertRaises(TypeError, self.assertAlmostRelativeEqual,
            1.0 | units.m, 1.0, places=3, expected_message="Cannot compare quantity: 1.0 m with non-quantity: 1.0.")

        self.assertRaises(TypeError, self.assertAlmostRelativeEqual,
            1.0, 1.0 | units.m, places=3, expected_message="Cannot compare non-quantity: 1.0 with quantity: 1.0 m.")
